from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator

import redis.asyncio as redis
from fastapi import Depends

from moriarty.matrix.operator_.config import Config, get_config


async def get_redis_client(
    config: Config = Depends(get_config),
) -> AsyncGenerator[Any, redis.Redis | redis.RedisCluster]:
    redis_url = config.redis.get_redis_url()
    if config.redis.cluster:
        redis_client = redis.RedisCluster.from_url(redis_url, decode_responses=True)
    else:
        redis_client = redis.from_url(redis_url, decode_responses=True)
    try:
        yield redis_client
    finally:
        await redis_client.aclose()


@asynccontextmanager
async def open_redis_client(
    config: Config,
) -> AsyncGenerator[Any, redis.Redis | redis.RedisCluster]:
    redis_url = config.redis.get_redis_url()
    if config.redis.cluster:
        redis_client = redis.RedisCluster.from_url(redis_url, decode_responses=True)
    else:
        redis_client = redis.from_url(redis_url, decode_responses=True)
    try:
        yield redis_client
    finally:
        await redis_client.aclose()



from pprint import pformat

import click

from .job_manager.bridge.manager import BridgeManager
from .operator_.spawner.manager import SpawnerManager


@click.command()
def list_plugins():
    click.echo(f"Registered Bridges: \n{pformat(BridgeManager().registed_cls)}")
    click.echo(f"Registered Spawners: \n{pformat(SpawnerManager().registed_cls)}")


@click.group()
def cli():
    pass


cli.add_command(list_plugins)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional

from pydantic import BaseModel

if TYPE_CHECKING:
    from moriarty.matrix.job_manager.params import InferenceJob


class InvokeParams(BaseModel):
    inference_id: Optional[str] = None
    endpoint_name: str
    invoke_params: dict
    priority: Optional[int] = None
    """
    1-100, Higher number means higher priority
    """

    created_at: datetime = datetime.now()
    expires_at: Optional[datetime] = None

    def model_dump_metadata(self) -> dict:
        return self.model_dump(
            exclude={"invoke_params"},
        )

    @classmethod
    def from_inference_job(cls, job: "InferenceJob") -> InvokeParams:
        return cls(
            inference_id=job.inference_id,
            endpoint_name=job.endpoint_name,
            invoke_params=job.payload,
            **job.metadata,
        )


class InvokeResponse(BaseModel):
    inference_id: str



import datetime
import os
import string
from typing import Optional

import escapism
from kubernetes_asyncio import client
from kubernetes_asyncio.client.api_client import ApiClient

from moriarty.__init__ import __version__
from moriarty.log import logger
from moriarty.matrix.operator_.orm import EndpointORM
from moriarty.matrix.operator_.params import EndpointRuntimeInfo
from moriarty.matrix.operator_.spawner.plugin import (
    EnvironmentBuilder,
    Spawner,
    hookimpl,
)
from moriarty.matrix.operator_.tools import load_kube_config


def get_current_namespace() -> str | None:
    if not os.path.exists("/var/run/secrets/kubernetes.io/serviceaccount/namespace"):
        return None
    return open("/var/run/secrets/kubernetes.io/serviceaccount/namespace").read()


class DeploymentMixin(EnvironmentBuilder):
    namespace: str
    init_image = os.getenv("INIT_IMAGE", "peakcom/s5cmd")
    sidecar_image = os.getenv("SIDECAR_IMAGE", "wh1isper/moriarty-sidecar")
    image_pull_secrets = os.getenv("IMAGE_PULL_SECRETS")

    def _escape_string(self, s: str) -> str:
        safe_chars = set(string.ascii_lowercase + string.digits)
        return escapism.escape(s, safe=safe_chars, escape_char="-", allow_collisions=True).lower()

    def get_deployment_name(self, endpoint_name: str) -> str:
        # Escape the endpoint name
        return self._escape_string(f"endpoint-{endpoint_name}")

    async def inspect_deployment_status(
        self, endpoint_name: str
    ) -> client.V1DeploymentStatus | None:
        deployment_name = self.get_deployment_name(endpoint_name)
        logger.debug(f"Inspect deployment status: {deployment_name}(as endpoint {endpoint_name}).")

        async with ApiClient() as api:
            v1 = client.AppsV1Api(api)

            try:
                deployment = await v1.read_namespaced_deployment_status(
                    name=deployment_name, namespace=self.namespace
                )
            except client.rest.ApiException as e:
                if e.status == 404:
                    return None
                raise

            return deployment.status

    async def make_and_apply_deployment(self, endpoint_orm: EndpointORM) -> None:
        deployment = self._make_deployment(endpoint_orm)
        try:
            await self._apply_deployment(deployment)
        except client.rest.ApiException as e:
            logger.error(
                f"Apply deployment failed: {e}. Input: {client.ApiClient().sanitize_for_serialization(deployment)}"
            )
            raise

    async def make_and_update_deployment(self, endpoint_orm: EndpointORM) -> None:
        deployment = self._make_deployment(endpoint_orm)
        try:
            await self._replace_deployment(deployment)
        except client.rest.ApiException as e:
            logger.error(
                f"Update deployment failed: {e}. Input: {client.ApiClient().sanitize_for_serialization(deployment)}"
            )
            raise

    async def delete_deployment(self, endpoint_name: str) -> None:
        deployment_name = self.get_deployment_name(endpoint_name)
        async with ApiClient() as api:
            v1 = client.AppsV1Api(api)
            try:
                await v1.delete_namespaced_deployment(
                    name=deployment_name, namespace=self.namespace
                )
            except client.rest.ApiException as e:
                if e.status == 404:
                    return
                raise

    async def restart_deployment(self, endpoint_name: str) -> None:
        now = datetime.datetime.now()
        now = str(now.isoformat("T") + "Z")
        body = {
            "spec": {
                "template": {
                    "metadata": {"annotations": {"kubectl.kubernetes.io/restartedAt": now}}
                }
            }
        }
        deployment = self.get_deployment_name(endpoint_name)
        namespace = self.namespace

        async with ApiClient() as api:
            v1_apps = client.AppsV1Api(api)
            await v1_apps.patch_namespaced_deployment(deployment, namespace, body, pretty="true")

    async def scale_deployment(self, endpoint_name: str, target_replicas: int) -> None:
        deployment = self.get_deployment_name(endpoint_name)
        namespace = self.namespace
        async with ApiClient() as api:
            v1 = client.AppsV1Api(api)
            await v1.patch_namespaced_deployment_scale(
                name=deployment,
                namespace=namespace,
                body={"spec": {"replicas": target_replicas}},
            )

    async def _apply_deployment(self, deployment: client.V1Deployment) -> None:
        async with ApiClient() as api:
            v1 = client.AppsV1Api(api)
            await v1.create_namespaced_deployment(namespace=self.namespace, body=deployment)

    async def _replace_deployment(self, deployment: client.V1Deployment) -> None:
        async with ApiClient() as api:
            v1 = client.AppsV1Api(api)
            await v1.replace_namespaced_deployment(
                name=deployment.metadata.name, namespace=self.namespace, body=deployment
            )

    def _make_deployment(
        self,
        endpoint_orm: EndpointORM,
    ) -> client.V1Deployment:
        deployment_name = self.get_deployment_name(endpoint_orm.endpoint_name)

        deployment = client.V1Deployment()

        deployment.api_version = "apps/v1"
        deployment.kind = "Deployment"

        deployment.metadata = client.V1ObjectMeta()
        deployment.metadata.name = deployment_name
        deployment.metadata.namespace = self.namespace
        deployment.metadata.labels = {
            "moriarty.version": __version__,
            "moriarty.kubespawner.version": __version__,
            "moriarty.kubespawner.gpu_nums": str(endpoint_orm.gpu_nums or 0),
            "moriarty.kubespawner.gpu_type": str(endpoint_orm.gpu_type or "").replace("/", "-"),
        }

        deployment.spec = client.V1DeploymentSpec(
            replicas=endpoint_orm.replicas,
            selector=client.V1LabelSelector(
                match_labels={
                    "app": "moriarty",
                    "moriarty.endpoint.deployment": deployment_name,
                }
            ),
            template=client.V1PodTemplateSpec(
                metadata=client.V1ObjectMeta(
                    labels={
                        "app": "moriarty",
                        "moriarty.endpoint.deployment": deployment_name,
                    },
                ),
                spec=client.V1PodSpec(
                    restart_policy="Always",
                    affinity=self._make_affinity(endpoint_orm),
                    node_selector=(
                        client.V1NodeSelector(
                            match_labels={k: v for k, v in endpoint_orm.node_labels.items() if v}
                        )
                        if endpoint_orm.node_labels
                        else None
                    ),
                    init_containers=[self._make_init_container(endpoint_orm)],
                    containers=[
                        self._make_sidecar_container(endpoint_orm),
                        self._make_compute_contaienr(endpoint_orm),
                    ],
                    volumes=[
                        client.V1Volume(
                            name="modeldir",
                            empty_dir=client.V1EmptyDirVolumeSource(),
                        ),
                    ],
                    image_pull_secrets=(
                        [
                            client.V1LocalObjectReference(
                                name=self.image_pull_secrets,
                            )
                        ]
                        if self.image_pull_secrets
                        else None
                    ),
                ),
            ),
        )

        return deployment

    def _make_affinity(self, endpoint_orm: EndpointORM) -> client.V1Affinity | None:
        if not endpoint_orm.node_affinity:
            return None
        return client.V1Affinity(
            node_affinity=client.V1NodeAffinity(
                required_during_scheduling_ignored_during_execution=client.V1NodeSelector(
                    node_selector_terms=[
                        client.V1NodeSelectorTerm(
                            match_expressions=[
                                {
                                    "key": k,
                                    "operator": "In",
                                    "values": v,
                                }
                                for k, v in endpoint_orm.node_affinity.items()
                            ]
                        )
                    ]
                )
            )
        )

    def _make_init_container(
        self,
        endpoint_orm: EndpointORM,
    ) -> client.V1Container:
        # Use init container to /opt/ml/model
        command = [
            "mkdir",
            "-p",
            "/opt/ml/model",
        ]

        if endpoint_orm.model_path:
            command.extend(
                [
                    "&&",
                    "/s5cmd",
                    "cp",
                    (
                        f"{endpoint_orm.model_path}*"
                        if endpoint_orm.model_path.endswith("/")
                        else endpoint_orm.model_path
                    ),
                    "/opt/ml/model/",
                ]
            )

        # Make sure the model dir is read-only
        command.extend(
            [
                "&&",
                "chmod",
                "-R",
                "444",
                "/opt/ml/model",
            ]
        )
        return client.V1Container(
            name="init",
            image=self.init_image,
            command=["/bin/sh"],
            args=["-c", " ".join(command)],
            volume_mounts=[
                client.V1VolumeMount(
                    name="modeldir",
                    mount_path="/opt/ml/model",
                ),
            ],
            env=[
                client.V1EnvVar(name=k, value=v) for k, v in self.build_init_environment().items()
            ],
        )

    def _make_sidecar_container(
        self,
        endpoint_orm: EndpointORM,
    ) -> client.V1Container:
        return client.V1Container(
            name="sidecar",
            image=self.sidecar_image,
            resources=client.V1ResourceRequirements(
                requests={
                    "cpu": "100m",
                    "memory": "256Mi",
                },
                limits={
                    "cpu": "100m",
                    "memory": "256Mi",
                },
            ),
            env=[
                client.V1EnvVar(name=k, value=v)
                for k, v in self.build_sidecar_environment(endpoint_orm).items()
            ],
        )

    def _make_compute_contaienr(
        self,
        endpoint_orm: EndpointORM,
    ) -> client.V1Container:
        cpu_limit = {}
        memory_limit = {}
        gpu_limit = {}
        if endpoint_orm.gpu_nums:
            gpu_limit = {
                endpoint_orm.gpu_type: f"{endpoint_orm.gpu_nums}",
            }
        if endpoint_orm.cpu_limit:
            cpu_limit = {
                "cpu": f"{endpoint_orm.cpu_limit}",
            }
        if endpoint_orm.memory_limit:
            memory_limit = {
                "memory": f"{endpoint_orm.memory_limit}Mi",
            }

        return client.V1Container(
            name="compute",
            image=endpoint_orm.image,
            command=endpoint_orm.commands,
            args=endpoint_orm.args,
            resources=client.V1ResourceRequirements(
                requests={
                    "cpu": f"{endpoint_orm.cpu_request}",
                    "memory": f"{endpoint_orm.memory_request}Mi",
                },
                limits={
                    **cpu_limit,
                    **memory_limit,
                    **gpu_limit,
                },
            ),
            env=[
                client.V1EnvVar(name=k, value=v)
                for k, v in self.build_compute_environment(endpoint_orm).items()
            ],
            env_from=[
                client.V1EnvFromSource(secret_ref=client.V1SecretEnvSource(name=secret_ref))
                for secret_ref in endpoint_orm.environment_variables_secret_refs
            ],
            ports=[
                client.V1ContainerPort(
                    name="http",
                    container_port=endpoint_orm.invoke_port,
                ),
            ],
            readiness_probe=client.V1Probe(
                http_get=client.V1HTTPGetAction(
                    path=endpoint_orm.health_check_path,
                    port=endpoint_orm.invoke_port,
                ),
                initial_delay_seconds=endpoint_orm.health_check_interval,
                period_seconds=endpoint_orm.health_check_interval,
                timeout_seconds=endpoint_orm.health_check_timeout,
                failure_threshold=10,
                success_threshold=1,
            ),
            volume_mounts=[
                client.V1VolumeMount(
                    name="modeldir",
                    mount_path="/opt/ml/model",
                ),
            ],
        )


class KubeSpawner(Spawner, DeploymentMixin):
    register_name = "kube"

    def __init__(
        self,
        namespace: Optional[str] = os.getenv("MORIARTY_KUBE_NAMESPACE_ENV"),
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.namespace = namespace or get_current_namespace() or "moriarty"

    async def prepare(self) -> None:
        await load_kube_config()

    async def count_avaliable_instances(self, endpoint_name: str) -> int:
        status = await self.inspect_deployment_status(endpoint_name)
        if status is None:
            logger.warning(f"Deployment not found: {endpoint_name}.")
            return 0

        return status.ready_replicas or 0

    async def create(self, endpoint_orm: EndpointORM) -> None:
        logger.info(f"Create deployment: {endpoint_orm.endpoint_name}.")
        await self.make_and_apply_deployment(endpoint_orm)

    async def update(self, endpoint_orm: EndpointORM, need_restart: bool = True) -> None:
        logger.info(f"Update deployment: {endpoint_orm.endpoint_name}.")
        await self.make_and_update_deployment(endpoint_orm)
        if need_restart:
            logger.info(f"Restart deployment: {endpoint_orm.endpoint_name}.")
            await self.restart_deployment(endpoint_orm.endpoint_name)

    async def delete(self, endpoint_name: str) -> None:
        logger.info(f"Delete deployment: {endpoint_name}.")
        await self.delete_deployment(endpoint_name)

    async def get_runtime_info(self, endpoint_name: str) -> EndpointRuntimeInfo:
        status = await self.inspect_deployment_status(endpoint_name)
        if not status:
            logger.warning(f"Deployment not found: {endpoint_name}.")
            return EndpointRuntimeInfo(
                total_replicas_nums=0,
                updated_replicas_nums=0,
                avaliable_replicas_nums=0,
                unavailable_replicas_nums=0,
            )

        return EndpointRuntimeInfo(
            total_replicas_nums=status.replicas or 0,
            updated_replicas_nums=status.updated_replicas or 0,
            avaliable_replicas_nums=status.ready_replicas or 0,
            unavailable_replicas_nums=status.unavailable_replicas or 0,
        )

    async def scale(self, endpoint_name: str, target_replicas: int) -> None:
        logger.info(f"Scale deployment: {endpoint_name}.")
        await self.scale_deployment(endpoint_name, target_replicas)


@hookimpl
def register(manager):
    manager.register(KubeSpawner)



import os
import uuid

from fastapi import Depends

from moriarty.envs import get_bridge_name
from moriarty.log import logger
from moriarty.matrix.connector.params import InvokeParams, InvokeResponse
from moriarty.matrix.job_manager.bridge_wrapper import BridgeWrapper, get_bridge_wrapper
from moriarty.matrix.job_manager.params import InferenceJob


def get_invoker(
    bridge_name: str = Depends(get_bridge_name),
    bridge_wrapper: BridgeWrapper = Depends(get_bridge_wrapper),
):
    return Invoker(bridge_name=bridge_name, bridge_wrapper=bridge_wrapper)


class Invoker:
    def __init__(
        self,
        bridge_name: str,
        bridge_wrapper: BridgeWrapper,
    ):
        self.bridge_name = bridge_name
        self.bridge_wrapper = bridge_wrapper

    async def invoke(self, params: InvokeParams) -> InvokeResponse:
        inference_id = params.inference_id or uuid.uuid4().hex
        logger.debug(f"Invoke inference: {inference_id} with params: {params}")
        await self.bridge_wrapper.enqueue_job(
            bridge=self.bridge_name,
            endpoint_name=params.endpoint_name,
            job=InferenceJob(
                inference_id=inference_id,
                payload={
                    **params.invoke_params,
                    "inference_id": inference_id,  # Patch inference_id in payload
                },
                metadata=params.model_dump_metadata(),
            ),
            priority=params.priority,
        )
        return InvokeResponse(inference_id=inference_id)



from __future__ import annotations

import typing

import kubernetes_asyncio as kubernetes
import pytest
from sqlalchemy import delete

from moriarty.matrix.operator_.orm import EndpointORM
from moriarty.matrix.operator_.spawner.manager import get_spawner, get_spawner_manager

if typing.TYPE_CHECKING:
    from moriarty.matrix.operator_.spawner.impl.k8s import KubeSpawner


@pytest.fixture
async def kube_spawner():
    try:
        return await get_spawner("kube", get_spawner_manager())
    except kubernetes.config.config_exception.ConfigException:
        pytest.skip("Kubernetes is not available, use minikube or kind for tests")


@pytest.fixture
def mock_endpoint_orm(session):
    endpoint_name = "mock_endpoint"
    orm = EndpointORM(
        endpoint_name=endpoint_name,
    )
    session.add(orm)
    session.commit()
    session.refresh(orm)
    yield orm
    session.execute(delete(EndpointORM).where(EndpointORM.endpoint_name == endpoint_name))
    session.commit()


async def test_kube_spawner(kube_spawner: KubeSpawner, mock_endpoint_orm):
    assert await kube_spawner.count_avaliable_instances(mock_endpoint_orm.endpoint_name) == 0

    await kube_spawner.create(mock_endpoint_orm)

    mock_endpoint_orm.cpu_request = 1
    await kube_spawner.update(mock_endpoint_orm)
    await kube_spawner.count_avaliable_instances(mock_endpoint_orm.endpoint_name)
    # runtime_info = await kube_spawner.get_runtime_info(mock_endpoint_orm.endpoint_name)
    # assert runtime_info.total_replicas_nums == 1

    await kube_spawner.scale(mock_endpoint_orm.endpoint_name, 2)
    # runtime_info = await kube_spawner.get_runtime_info(mock_endpoint_orm.endpoint_name)
    # assert runtime_info.total_replicas_nums == 2

    await kube_spawner.delete(mock_endpoint_orm.endpoint_name)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from moriarty.matrix.job_manager.bridge.plugin import QueueBridge, hookimpl


class CustomBridge(QueueBridge):
    register_name = "example"


@hookimpl
def register(manager):
    manager.register(CustomBridge)



from __future__ import annotations

import enum
import traceback
from typing import Optional

import httpx
from pydantic import BaseModel


class InferenceProxyStatus(str, enum.Enum):
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    TIMEOUT = "TIMEOUT"
    INTERNAL_ERROR = "INTERNAL_ERROR"


class MatrixCallback(BaseModel):
    inference_id: str
    status: InferenceProxyStatus
    msg: Optional[str] = None
    payload: Optional[str] = None

    @classmethod
    def from_exception(cls, inference_id: str, e: Exception) -> MatrixCallback:
        trb = "".join(traceback.format_tb(e.__traceback__))
        if isinstance(e, httpx.TimeoutException):
            return MatrixCallback(
                inference_id=inference_id,
                status=InferenceProxyStatus.TIMEOUT,
                msg=f"Inference {inference_id} timeout",
                payload=trb,
            )
        else:
            return MatrixCallback(
                inference_id=inference_id,
                status=InferenceProxyStatus.INTERNAL_ERROR,
                msg=f"Inference {inference_id} failed: {e}",
                payload=trb,
            )

    @classmethod
    def from_response(cls, inference_id: str, response: httpx.Response) -> MatrixCallback:
        return MatrixCallback(
            inference_id=inference_id,
            status=InferenceProxyStatus.FINISHED,
            msg=f"Inference {inference_id} finished",
            payload=response.text,
        )



from __future__ import annotations

import asyncio
import base64
import json
from typing import Awaitable

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

from fastapi.concurrency import run_in_threadpool

from moriarty.log import logger
from moriarty.matrix.job_manager.bridge.plugin import QueueBridge, hookimpl
from moriarty.matrix.job_manager.params import (
    InferenceJob,
    InferenceResult,
    InferenceResultStatus,
)


class SQSBridge(QueueBridge):
    """
    Bridge to SQS. Compatible with AWS SageMaker AsyncInference.

    Will create SQS queue for each endpoint.

    Args:
        bridge_result_queue_url: AWS SQS URL for result queue
        poll_time: polling time in seconds
    """

    register_name = "sqs"

    def __init__(
        self,
        bridge_result_queue_url: str = None,
        poll_time: int = 1,
        *args,
        **kwargs,
    ):
        super().__init__(bridge_result_queue_url, *args, **kwargs)
        self.poll_time = poll_time

    @property
    def client(self):
        try:
            import boto3
        except ImportError:
            raise RuntimeError(
                "boto3 is not installed, try install moriarty with `pip install moriarty[matrix]` for all components or `pip install moriarty[sqs]` for sqs only"
            )
        return boto3.client("sqs")

    def _get_queue_url_prefix(self, endpoint_name: str) -> str:
        return f"moriarty-bridge-{endpoint_name}"

    def make_queue_name(self, endpoint_name: str, priority: int = None) -> str:
        priority = priority or 100
        return f"{self._get_queue_url_prefix(endpoint_name)}-{priority}"

    async def list_avaliable_priorities(self, endpoint_name: str) -> list[int]:
        def _():
            priorities = []
            response = self.client.list_queues(
                QueueNamePrefix=self._get_queue_url_prefix(endpoint_name)
            )
            while response["QueueUrls"]:
                for url in response["QueueUrls"]:
                    response = self.client.get_queue_attributes(
                        QueueUrl=url, AttributeNames=["ApproximateNumberOfMessages"]
                    )
                    if int(response["Attributes"]["ApproximateNumberOfMessages"]) == 0:
                        continue

                    queue_name = url.split("/")[-1]
                    priority = int(queue_name.split("-")[-1])
                    priorities.append(priority)
                if "NextToken" in response:
                    response = self.client.list_queues(
                        QueueNamePrefix=self._get_queue_url_prefix(endpoint_name),
                        NextToken=response["NextToken"],
                    )
                else:
                    break
            return sorted(priorities)

        return await run_in_threadpool(_)

    @cache
    def make_job_queue_url(self, endpoint_name: str, priority: int = None) -> str:
        queue_name = self.make_queue_name(endpoint_name, priority)
        try:
            return self.client.get_queue_url(QueueName=queue_name)["QueueUrl"]
        except self.client.exceptions.QueueDoesNotExist:
            logger.info(f"Create queue: {queue_name} as it does not exist")
            response = self.client.create_queue(QueueName=queue_name)
            return response["QueueUrl"]

    def remove_job_queue(self, endpoint_name: str, priority: int = None) -> None:
        queue_name = self.make_queue_name(endpoint_name, priority)
        self.client.delete_queue(QueueUrl=queue_name)
        logger.info(f"Queue {queue_name} removed")

    async def enqueue_job(self, endpoint_name: str, job: InferenceJob, priority: int = None) -> str:
        queue_url = self.make_job_queue_url(endpoint_name, priority=priority)
        await run_in_threadpool(
            self.client.send_message,
            QueueUrl=queue_url,
            MessageBody=job.model_dump_json(),
        )
        logger.debug(f"Job {job.inference_id} enqueued")
        return job.inference_id

    async def dequeue_job(
        self,
        endpoint_name: str,
        process_func: Awaitable[InferenceJob],
        size: int = 1,
        priority: int = None,
    ) -> int:
        url = self.make_job_queue_url(endpoint_name, priority)
        logger.debug(f"Polling job from queue: {url}")
        return await self._poll_job_and_execute(process_func, url, size)

    async def enqueue_result(self, result: InferenceResult) -> str:
        invocation_status = (
            "Completed" if result.status == InferenceResultStatus.COMPLETED else "Failed"
        )
        response_body = (
            {
                "content": base64.b64encode(json.dumps(result.payload).encode()).decode(),
                "message": result.message,
                "encoding": "BASE64",
            }
            if result.status == InferenceResultStatus.COMPLETED
            else {
                "content": result.payload,
                "message": result.message,
            }
        )

        message = {
            "Message": json.dumps(
                {
                    "invocationStatus": invocation_status,
                    "eventName": "InferenceResult",
                    "responseBody": response_body,
                    "inferenceId": result.inference_id,
                    "eventName": "InferenceResult",
                }
            )
        }

        await run_in_threadpool(
            self.client.send_message,
            QueueUrl=self.bridge_result_queue_url,
            MessageBody=json.dumps(message),
        )
        logger.debug(f"Result {result.inference_id} enqueued")
        return result.inference_id

    async def dequeue_result(self, process_func: Awaitable[InferenceResult], size: int = 10) -> int:
        logger.debug(f"Polling result from queue: {self.bridge_result_queue_url}")
        return await self._poll_result_and_execute(process_func, self.bridge_result_queue_url, size)

    async def _poll_job_and_execute(self, process_func: Awaitable[InferenceJob], url, size) -> int:
        if size and size > 10:
            messages = []
            while len(messages) < size:
                messages_len_before = len(messages)
                messages.extend(await self._poll_message(url, 10))
                if len(messages) == messages_len_before:
                    # No message left
                    break
        else:
            messages = await self._poll_job_message(url, size)

        await asyncio.gather(
            *[
                self._wrap_process_func(process_func, job, url, receipt_handle)
                for job, receipt_handle in messages
            ]
        )
        return len(messages)

    async def _poll_job_message(self, url, size: int = None) -> list[tuple[InferenceJob, str]]:
        size = size or 1
        valid_messages = []
        response = await run_in_threadpool(
            self.client.receive_message,
            QueueUrl=url,
            MaxNumberOfMessages=size,
            WaitTimeSeconds=self.poll_time,
        )
        if "Messages" not in response:
            return valid_messages

        for message in response["Messages"]:
            receipt_handle = message["ReceiptHandle"]
            try:
                body = message["Body"]
                message = InferenceJob.model_validate_json(body)
            except Exception as e:
                logger.error(f"Error parsing message {body}: {e}")
                logger.exception(e)
            else:
                valid_messages.append((message, receipt_handle))
        return valid_messages

    async def _poll_result_and_execute(
        self, process_func: Awaitable[InferenceResult], url, size
    ) -> int:
        if size and size > 10:
            messages = []
            while len(messages) < size:
                messages_len_before = len(messages)
                messages.extend(await self._poll_message(url, 10))
                if len(messages) == messages_len_before:
                    # No message left
                    break
        else:
            messages = await self._poll_result_message(url, size)

        logger.debug(f"Got {len(messages)} messages")
        for result, receipt_handle in messages:
            await self._wrap_process_func(
                process_func, result, self.bridge_result_queue_url, receipt_handle
            )
        return len(messages)

    async def _poll_result_message(
        self, url, size: int = None
    ) -> list[tuple[InferenceResult, str]]:
        size = size or 1
        valid_messages = []
        response = await run_in_threadpool(
            self.client.receive_message,
            QueueUrl=url,
            MaxNumberOfMessages=size,
            WaitTimeSeconds=self.poll_time,
        )
        if "Messages" not in response:
            return valid_messages

        for message in response["Messages"]:
            receipt_handle = message["ReceiptHandle"]
            try:
                message_body = json.loads(message["Body"])
                message = json.loads(message_body["Message"])
            except Exception as e:
                logger.error(f"Error parsing message {message}: {e}")
                logger.exception(e)
                continue
            if "eventName" in message and message["eventName"] != "InferenceResult":
                logger.warning(f"Unexpected message {message}. Message will be ignored.")
                continue
            if "invocationStatus" in message and message["invocationStatus"] == "Failed":
                logger.debug(f"Result {message['inferenceId']} failed")
                valid_messages.append(
                    (
                        InferenceResult(
                            inference_id=message["inferenceId"],
                            status=InferenceResultStatus.FAILED,
                            message=message["responseBody"]["message"],
                            payload=message["responseBody"]["content"],
                        ),
                        receipt_handle,
                    )
                )
                continue
            try:
                logger.debug(f"Result {message['inferenceId']} completed")
                inference_result = InferenceResult(
                    inference_id=message["inferenceId"],
                    status=InferenceResultStatus.COMPLETED,
                    message=message["responseBody"]["message"],
                    payload=json.loads(
                        base64.b64decode(message["responseBody"]["content"]).decode("utf-8")
                    ),
                )
                valid_messages.append(
                    (
                        inference_result,
                        receipt_handle,
                    )
                )
                continue
            except Exception as e:
                logger.error(f"Cannot parse inference result {message}.")
                logger.exception(e)
                continue

        return valid_messages

    async def _wrap_process_func(
        self,
        process_func: Awaitable[InferenceJob] | Awaitable[InferenceResult],
        job: InferenceJob,
        url: str,
        receipt_handle: str,
    ):
        try:
            await process_func(job)
        except Exception as e:
            logger.error(f"Error processing job {job.inference_id}: {e}")
            logger.exception(e)
            raise
        else:
            await run_in_threadpool(
                self.client.delete_message,
                QueueUrl=url,
                ReceiptHandle=receipt_handle,
            )


@hookimpl
def register(manager):
    manager.register(SQSBridge)



from moriarty.matrix.operator_.spawner.manager import SpawnerManager


def test_spawner_manager():
    registed = ["kube"]
    manager = SpawnerManager()

    assert sorted(registed) == sorted(manager.registed_cls.keys())



from moriarty.sidecar.params import InferenceProxyStatus, MatrixCallback


def test_hello(client):
    response = client.get("/")
    assert response.status_code == 200


def test_callback(client, inference_log):
    response = client.post(
        "/callback",
        data=MatrixCallback(
            inference_id=inference_log,
            status=InferenceProxyStatus.FINISHED,
            msg=f"Inference {inference_log} finished",
            payload="{}",
        ).model_dump_json(),
    )
    assert response.status_code == 200



from custom_bridge.imp import CustomBridge

from moriarty.matrix.job_manager.bridge.manager import BridgeManager


def test_registed():
    manager = BridgeManager()

    assert CustomBridge.register_name in manager.registed_cls.keys()



from __future__ import annotations

import os
from functools import cached_property

import redis.asyncio as redis
from fastapi import Depends, HTTPException, status
from sqlalchemy import delete, func, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from moriarty.envs import get_bridge_result_queue_url
from moriarty.log import logger
from moriarty.matrix.connector.invoker import get_bridge_name
from moriarty.matrix.job_manager.bridge_wrapper import BridgeWrapper, get_bridge_wrapper
from moriarty.matrix.job_manager.params import InferenceJob, InferenceResult
from moriarty.matrix.operator_.autoscaler import (
    AutoscalerManager,
    get_autoscaler_manager,
)
from moriarty.matrix.operator_.dbutils import get_db_session
from moriarty.matrix.operator_.mixin import (
    AutoscaleMixin,
    EndpointMixin,
    MetricsManager,
)
from moriarty.matrix.operator_.orm import AutoscalerORM, EndpointORM, InferenceLogORM
from moriarty.matrix.operator_.params import (
    ContainerScope,
    CreateEndpointParams,
    ListEndpointsResponse,
    QueryEndpointResponse,
    ResourceScope,
    ScheduleScope,
    UpdateEndpointParams,
)
from moriarty.matrix.operator_.rds import get_redis_client
from moriarty.matrix.operator_.spawner import plugin
from moriarty.matrix.operator_.spawner.manager import get_spawner
from moriarty.sidecar.params import MatrixCallback
from moriarty.sidecar.producer import JobProducer


def get_bridger(
    spawner: plugin.Spawner = Depends(get_spawner),
    bridge_name: str = Depends(get_bridge_name),
    bridge_wrapper: BridgeWrapper = Depends(get_bridge_wrapper),
    redis_client: redis.Redis | redis.RedisCluster = Depends(get_redis_client),
    session: AsyncSession = Depends(get_db_session),
    bridge_result_queue_url: str = Depends(get_bridge_result_queue_url),
) -> Bridger:
    return Bridger(
        spawner=spawner,
        bridge_name=bridge_name,
        bridge_wrapper=bridge_wrapper,
        redis_client=redis_client,
        session=session,
        bridge_result_queue_url=bridge_result_queue_url,
    )


async def get_operaotr(
    spawner: plugin.Spawner = Depends(get_spawner),
    bridger: Bridger = Depends(get_bridger),
    session: AsyncSession = Depends(get_db_session),
    redis_client: redis.Redis | redis.RedisCluster = Depends(get_redis_client),
    autoscaler_manager: AutoscalerManager = Depends(get_autoscaler_manager),
) -> Operator:
    return Operator(
        spawner=spawner,
        bridger=bridger,
        session=session,
        redis_client=redis_client,
        autoscaler_manager=autoscaler_manager,
    )


class Bridger(EndpointMixin, AutoscaleMixin):
    def __init__(
        self,
        spawner: plugin.Spawner,
        bridge_name: str,
        bridge_wrapper: BridgeWrapper,
        redis_client: redis.Redis | redis.RedisCluster,
        session: AsyncSession,
        bridge_result_queue_url: None | str = None,
    ) -> None:
        self.spawner = spawner
        self.bridge_name = bridge_name
        self.bridge_wrapper = bridge_wrapper
        self.redis_client = redis_client
        self.session = session
        self.bridge_result_queue_url = bridge_result_queue_url

        self.metrics_manager = MetricsManager(
            spawner=spawner,
            redis_client=self.redis_client,
        )

    @cached_property
    def job_producer(self) -> JobProducer:
        return JobProducer(redis_client=self.redis_client)

    async def bridge_all(self) -> None:
        for endpoint_name in await self.get_avaliable_endpoints():
            await self.bridge_one(endpoint_name)

    async def has_capacity(self, endpoint_name: str) -> bool:
        endpoint_orm = await self.get_endpoint_orm(endpoint_name)
        if endpoint_orm is None:
            logger.warning(f"Endpoint not found: {endpoint_name}, may be deleted?")
            return False

        pending_count = await self.job_producer.count_unprocessed_jobs(endpoint_name)
        logger.debug(f"Pending count for {endpoint_name}: {pending_count}")

        autoscaler_orm = await self.get_autoscaler_orm(endpoint_name)
        if not autoscaler_orm:
            # No autoscaler, use default behavior
            concurrency = endpoint_orm.concurrency
            return (
                pending_count
                < endpoint_orm.queue_capacity
                + await self.spawner.count_avaliable_instances(endpoint_name) * concurrency
            )
        else:
            return pending_count < self.metrics_manager.calculate_queue_capacity(autoscaler_orm)

    async def bridge_one(self, endpoint_name: str) -> None:
        async def _warp_produce_job(job: InferenceJob) -> None:
            is_processed = (
                await self.session.execute(
                    select(InferenceLogORM).where(InferenceLogORM.inference_id == job.inference_id)
                )
            ).scalar_one_or_none() is not None
            if is_processed:
                logger.info(f"Job already processed: {job}, skip")
                return

            await self.job_producer.invoke(
                endpoint_name,
                params=job.payload,
            )
            log_orm = InferenceLogORM(
                inference_id=job.inference_id,
                endpoint_name=endpoint_name,
                inference_job=job.model_dump(),
            )
            self.session.add(log_orm)
            await self.session.commit()
            logger.debug(f"Inference job logged: {job}")

        logger.info(f"Bridge endpoint: {endpoint_name}")
        while await self.has_capacity(endpoint_name):
            sampled_count = await self.bridge_wrapper.sample_job(
                bridge=self.bridge_name,
                endpoint_name=endpoint_name,
                process_func=_warp_produce_job,
            )
            if not sampled_count:
                return
            logger.debug(f"One job sampled -> {endpoint_name}")

    async def bridge_result(self, callback: MatrixCallback) -> None:
        await self.bridge_wrapper.enqueue_result(
            bridge=self.bridge_name,
            bridge_result_queue_url=self.bridge_result_queue_url,
            result=InferenceResult.from_proxy_callback(callback),
        )


class Operator:
    def __init__(
        self,
        spawner: plugin.Spawner,
        bridger: Bridger,
        session: AsyncSession,
        redis_client: redis.Redis | redis.RedisCluster,
        autoscaler_manager: AutoscalerManager,
    ) -> None:
        self.spawner = spawner
        self.bridger = bridger
        self.session = session
        self.redis_client = redis_client
        self.autoscaler_manager = autoscaler_manager

    async def get_endpoint_orm(self, endpoint_name: str) -> EndpointORM | None:
        return (
            await self.session.execute(
                select(EndpointORM).where(EndpointORM.endpoint_name == endpoint_name)
            )
        ).scalar_one_or_none()

    async def get_endpoint_info(self, endpoint: str | EndpointORM) -> QueryEndpointResponse:
        if isinstance(endpoint, str):
            endpoint_orm = await self.get_endpoint_orm(endpoint)
        else:
            endpoint_orm = endpoint

        if not endpoint_orm:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Endpoint not found")

        endpoint_name = endpoint_orm.endpoint_name
        return QueryEndpointResponse(
            endpoint_name=endpoint_orm.endpoint_name,
            image=endpoint_orm.image,
            model_path=endpoint_orm.model_path,
            queue_capacity=endpoint_orm.queue_capacity,
            replicas=endpoint_orm.replicas,
            resource=ResourceScope(
                cpu_request=endpoint_orm.cpu_request,
                cpu_limit=endpoint_orm.cpu_limit,
                memory_request=endpoint_orm.memory_request,
                memory_limit=endpoint_orm.memory_limit,
                gpu_nums=endpoint_orm.gpu_nums,
                gpu_type=endpoint_orm.gpu_type,
            ),
            schedule=ScheduleScope(
                node_labels=endpoint_orm.node_labels,
                node_affinity=endpoint_orm.node_affinity,
            ),
            container=ContainerScope(
                environment_variables=endpoint_orm.environment_variables,
                environment_variables_secret_refs=endpoint_orm.environment_variables_secret_refs,
                commands=endpoint_orm.commands,
                args=endpoint_orm.args,
                invoke_port=endpoint_orm.invoke_port,
                invoke_path=endpoint_orm.invoke_path,
                health_check_path=endpoint_orm.health_check_path,
            ),
            runtime=await self.spawner.get_runtime_info(endpoint_name),
        )

    async def list_endpoints(
        self,
        limit: int,
        cursor: str | None,
        keyword: str | None,
        order_by: str,
        order: str,
    ) -> ListEndpointsResponse:
        if cursor:
            cursor = await self.get_endpoint_orm(cursor)
            if not cursor:
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND, detail="Cursor not found"
                )

        query = select(EndpointORM)
        if keyword:
            query = query.where(EndpointORM.endpoint_name.contains(keyword))

        if cursor:
            query = query.where(EndpointORM.id_ < cursor.id_)

        if order == "desc":
            query = query.order_by(getattr(EndpointORM, order_by).desc())

        else:
            query = query.order_by(getattr(EndpointORM, order_by))

        query = query.limit(limit)
        endpoint_orms = (await self.session.execute(query)).scalars().all()
        next_cursor = endpoint_orms[-1].endpoint_name if endpoint_orms else None

        return ListEndpointsResponse(
            endpoints=[
                await self.get_endpoint_info(endpoint_orm) for endpoint_orm in endpoint_orms
            ],
            cursor=next_cursor,
            limit=limit,
            total=(
                await self.session.execute(select(func.count(EndpointORM.endpoint_name)))
            ).scalar_one(),
        )

    async def create_endpoint(self, params: CreateEndpointParams) -> None:
        if await self.get_endpoint_orm(params.endpoint_name):
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT, detail="Endpoint already exists"
            )

        endpoint_orm = EndpointORM(
            endpoint_name=params.endpoint_name,
            queue_capacity=params.queue_capacity,
            image=params.image,
            model_path=params.model_path,
            replicas=params.replicas,
            **{
                k: v
                for k, v in dict(
                    cpu_request=params.resource.cpu_request,
                    cpu_limit=params.resource.cpu_limit,
                    memory_request=params.resource.memory_request,
                    memory_limit=params.resource.memory_limit,
                    gpu_nums=params.resource.gpu_nums,
                    gpu_type=params.resource.gpu_type,
                ).items()
                if v is not None
            },
            **{
                k: v
                for k, v in dict(
                    environment_variables=params.container.environment_variables,
                    environment_variables_secret_refs=params.container.environment_variables_secret_refs,
                    commands=params.container.commands,
                    args=params.container.args,
                    invoke_port=params.container.invoke_port,
                    invoke_path=params.container.invoke_path,
                    health_check_path=params.container.health_check_path,
                ).items()
                if v is not None
            },
            **{
                k: v
                for k, v in dict(
                    node_labels=params.schedule.node_labels,
                    node_affinity=params.schedule.node_affinity,
                ).items()
                if v is not None
            },
            **{
                k: v
                for k, v in dict(
                    concurrency=params.sidecar.concurrency,
                    process_timeout=params.sidecar.process_timeout,
                    health_check_timeout=params.sidecar.health_check_timeout,
                    health_check_interval=params.sidecar.health_check_interval,
                ).items()
                if v is not None
            },
        )
        self.session.add(endpoint_orm)
        await self.session.commit()
        await self.session.refresh(endpoint_orm)
        try:
            await self.spawner.create(endpoint_orm)
        except Exception as e:
            logger.exception(e)
            logger.info(f"Rollback create endpoint: {endpoint_orm.endpoint_name}")
            await self.session.delete(endpoint_orm)
            await self.session.commit()
            raise e

    async def update_endpoint(self, endpoint_name: str, params: UpdateEndpointParams) -> None:
        endpoint_orm = await self.get_endpoint_orm(endpoint_name)
        if not endpoint_orm:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Endpoint not found")

        if (
            params.replicas
            and endpoint_orm.replicas != params.replicas
            and await self.autoscaler_manager.get_autoscaler_orm(endpoint_name)
        ):
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail="Replicas cannot be updated as autoscaler is enabled. Try set autoscaler.",
            )

        await self.session.execute(
            update(EndpointORM)
            .where(EndpointORM.endpoint_name == endpoint_name)
            .values(
                **{
                    k: v
                    for k, v in dict(
                        queue_capacity=params.queue_capacity,
                        image=params.image,
                        model_path=params.model_path,
                        replicas=params.replicas,
                    ).items()
                    if v is not None
                },
                **(
                    {
                        k: v
                        for k, v in dict(
                            cpu_request=params.resource.cpu_request,
                            cpu_limit=params.resource.cpu_limit,
                            memory_request=params.resource.memory_request,
                            memory_limit=params.resource.memory_limit,
                            gpu_nums=params.resource.gpu_nums,
                            gpu_type=params.resource.gpu_type,
                        ).items()
                        if v is not None
                    }
                    if params.resource is not None
                    else {}
                ),
                **(
                    {
                        k: v
                        for k, v in dict(
                            environment_variables=params.container.environment_variables,
                            environment_variables_secret_refs=params.container.environment_variables_secret_refs,
                            commands=params.container.commands,
                            args=params.container.args,
                            invoke_port=params.container.invoke_port,
                            invoke_path=params.container.invoke_path,
                            health_check_path=params.container.health_check_path,
                        ).items()
                        if v is not None
                    }
                    if params.container is not None
                    else {}
                ),
                **(
                    {
                        k: v
                        for k, v in dict(
                            node_labels=params.schedule.node_labels,
                            node_affinity=params.schedule.node_affinity,
                        ).items()
                        if v is not None
                    }
                    if params.schedule is not None
                    else {}
                ),
                **(
                    {
                        k: v
                        for k, v in dict(
                            concurrency=params.sidecar.concurrency,
                            process_timeout=params.sidecar.process_timeout,
                            health_check_timeout=params.sidecar.health_check_timeout,
                            health_check_interval=params.sidecar.health_check_interval,
                        ).items()
                        if v is not None
                    }
                    if params.schedule is not None
                    else {}
                ),
            )
        )
        await self.session.commit()
        await self.session.refresh(endpoint_orm)
        try:
            await self.spawner.update(endpoint_orm, need_restart=params.need_restart)
        except Exception as e:
            logger.exception(e)
            logger.warning(f"Failed to update deployment but already update db. Try update again.")
            raise

        return await self.get_endpoint_info(endpoint_orm)

    async def delete_endpoint(self, endpoint_name: str) -> None:
        await self.spawner.delete(endpoint_name)
        await self.session.execute(
            delete(EndpointORM).where(EndpointORM.endpoint_name == endpoint_name)
        )
        await self.session.commit()
        await self.autoscaler_manager.delete(endpoint_name)

    async def handle_callback(self, callback: MatrixCallback) -> None:
        await self.bridger.bridge_result(callback)
        await self.session.execute(
            update(InferenceLogORM)
            .where(InferenceLogORM.inference_id == callback.inference_id)
            .values(
                status=callback.status,
                callback_response=callback.model_dump(),
                finished_at=func.now(),
            )
        )
        await self.session.commit()
        logger.info(f"Callback handled: {callback}")



import redis.asyncio as redis
from brq.models import Job
from brq.producer import Producer

from moriarty.log import logger

from .vars import GROUP_NAME, REDIS_PREFIX


class JobProducer:
    """
    A warpper of brq.producer.Producer
    """

    REDIS_PREFIX = REDIS_PREFIX
    GROUP_NAME = GROUP_NAME

    def __init__(self, redis_client: redis.Redis) -> None:
        self.redis_client = redis_client

    @property
    def producer(self):
        return Producer(self.redis_client, redis_prefix=self.REDIS_PREFIX)

    async def invoke(self, endpoint_name: str, params: dict) -> Job:
        logger.debug(f"Invoke endpoint: {endpoint_name}")
        return await self.producer.run_job(endpoint_name, kwargs={"payload": params})

    async def count_unprocessed_jobs(self, endpoint_name: str) -> int:
        return await self.producer.count_unprocessed_jobs(endpoint_name, group_name=self.GROUP_NAME)

    async def count_jobs(self, endpoint_name: str) -> int:
        return await self.producer.count_stream(endpoint_name)



import pytest
from sqlalchemy import delete

from moriarty.matrix.operator_.autoscaler import AutoscalerManager
from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.matrix.operator_.orm import AutoscaleLogORM, AutoscalerORM, EndpointORM


@pytest.fixture
async def autoscaler(
    async_redis_client,
    async_session,
    spawner,
):
    async def count_pending_jobs(endpoint_name: str) -> int:
        return 10

    autoscaler = AutoscalerManager(
        redis_client=async_redis_client, session=async_session, spawner=spawner
    )
    autoscaler.metrics_manager.count_pending_jobs = count_pending_jobs
    yield autoscaler


@pytest.fixture
def scaleable_endpoint(session):
    endpoint_name = "mock_scaleable_endpoint"

    session.add(
        EndpointORM(
            endpoint_name=endpoint_name,
            image="busybox",
        )
    )
    session.add(
        AutoscalerORM(
            endpoint_name=endpoint_name,
            min_replicas=1,
            max_replicas=10,
            scale_in_cooldown=60,
            scale_out_cooldown=60,
            metrics=MetricType.pending_jobs,
            metrics_threshold=3,
        )
    )
    session.commit()

    yield endpoint_name

    session.execute(delete(EndpointORM).where(EndpointORM.endpoint_name == endpoint_name))
    session.commit()

    session.execute(delete(AutoscalerORM).where(AutoscalerORM.endpoint_name == endpoint_name))
    session.commit()

    session.execute(delete(AutoscaleLogORM).where(AutoscaleLogORM.endpoint_name == endpoint_name))
    session.commit()


async def test_scan_and_scale(autoscaler: AutoscalerManager, scaleable_endpoint):
    await autoscaler.scan_and_scale()

    response = await autoscaler.get_autoscale_logs(
        scaleable_endpoint,
        limit=100,
        cursor=None,
        order_by="created_at",
        order="desc",
    )

    assert len(response.logs) == 1
    assert response.logs[0].old_replicas == 1
    assert response.logs[0].new_replicas == 7  # pending_jobs(10) - metrics_threshold(3) = 7


async def calculate_scale(autoscaler: AutoscalerManager):
    assert autoscaler.calculate_least_replicas(MetricType.pending_jobs, 5, 3) == 2
    assert autoscaler.calculate_least_replicas(MetricType.pending_jobs, 5, 0) == 5
    assert autoscaler.calculate_least_replicas(MetricType.pending_jobs, 5, 0.1) == 5
    assert autoscaler.calculate_least_replicas(MetricType.pending_jobs_per_instance, 10, 3) == 2
    assert autoscaler.calculate_least_replicas(MetricType.pending_jobs_per_instance, 10, 0) == 10



from datetime import datetime
from typing import Optional

from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.tools import FlexibleModel


class EndpointRuntimeInfo(FlexibleModel):
    total_replicas_nums: int
    updated_replicas_nums: int
    avaliable_replicas_nums: int
    unavailable_replicas_nums: int


class ResourceScope(FlexibleModel):
    cpu_request: Optional[float] = None
    cpu_limit: Optional[float] = None
    memory_request: Optional[float] = None
    memory_limit: Optional[float] = None
    gpu_nums: Optional[int] = None
    gpu_type: Optional[str] = None


class ScheduleScope(FlexibleModel):
    node_labels: Optional[dict] = None
    node_affinity: Optional[dict] = None


class ContainerScope(FlexibleModel):
    environment_variables: Optional[dict] = None
    environment_variables_secret_refs: Optional[list] = None
    commands: Optional[list] = None
    args: Optional[list] = None
    invoke_port: Optional[int] = 8080
    invoke_path: Optional[str] = "/invocations"
    health_check_path: Optional[str] = "/ping"


class SidecarScope(FlexibleModel):
    concurrency: Optional[int] = 1
    process_timeout: Optional[int] = 3600
    health_check_timeout: Optional[int] = 1200
    health_check_interval: Optional[int] = 5


class CreateEndpointParams(FlexibleModel):
    endpoint_name: str
    image: str
    model_path: Optional[str] = None

    queue_capacity: Optional[int] = 5
    replicas: int = 1
    resource: Optional[ResourceScope] = ResourceScope()
    schedule: Optional[ScheduleScope] = ScheduleScope()
    container: Optional[ContainerScope] = ContainerScope()
    sidecar: Optional[SidecarScope] = SidecarScope()


class QueryEndpointResponse(FlexibleModel):
    endpoint_name: str
    image: str
    model_path: Optional[str]
    queue_capacity: Optional[int]
    replicas: int
    resource: ResourceScope
    schedule: ScheduleScope
    container: ContainerScope
    runtime: EndpointRuntimeInfo


class ListEndpointsResponse(FlexibleModel):
    endpoints: list[QueryEndpointResponse]
    cursor: Optional[str] = None
    limit: int
    total: int


class UpdateEndpointParams(FlexibleModel):
    image: Optional[str] = None
    model_path: Optional[str] = None

    queue_capacity: Optional[int] = None
    replicas: Optional[int] = None

    resource: Optional[ResourceScope] = None
    schedule: Optional[ScheduleScope] = None
    container: Optional[ContainerScope] = None
    sidecar: Optional[SidecarScope] = None

    need_restart: bool = True


class QueryEndpointAutoscaleResponse(FlexibleModel):
    endpoint_name: str
    min_replicas: int
    max_replicas: int
    scale_in_cooldown: int
    scale_out_cooldown: int
    metrics: MetricType
    metrics_threshold: float


class SetAutoscaleParams(FlexibleModel):
    min_replicas: Optional[int] = None
    max_replicas: Optional[int] = None
    scale_in_cooldown: Optional[int] = None
    scale_out_cooldown: Optional[int] = None
    metrics: Optional[MetricType] = None
    metrics_threshold: Optional[float] = None


class AutoscaleLog(FlexibleModel):
    id_: int
    endpoint_name: str
    old_replicas: int
    new_replicas: int
    metrics: MetricType
    metrics_threshold: float
    metrics_log: dict
    details: Optional[str] = None
    created_at: datetime
    updated_at: datetime


class QueryEndpointAutoscaleLogResponse(FlexibleModel):
    logs: list[AutoscaleLog]
    cursor: Optional[int] = None
    limit: int
    total: int



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
import redis
from click.testing import CliRunner
from fastapi.testclient import TestClient

from moriarty import mock
from moriarty.envs import get_bridge_name
from moriarty.matrix.job_manager.bridge_wrapper import get_bridge_manager
from moriarty.matrix.operator_.api_app import app as APP
from moriarty.matrix.operator_.cli import drop
from moriarty.matrix.operator_.spawner.manager import (
    get_spawner_manager,
    get_spawner_name,
)


@pytest.fixture
async def app(pg_port, redis_port, monkeypatch, spawner_manager, bridge_manager):
    monkeypatch.setenv("REDIS_PORT", str(redis_port))
    monkeypatch.setenv("DB_PORT", str(pg_port))

    runner = CliRunner()
    # Drop all before testing
    result = runner.invoke(drop, ["--yes"])
    assert result.exit_code == 0

    # Clear redis data
    redis_client = redis.Redis(host="localhost", port=redis_port)
    redis_client.flushall()

    APP.dependency_overrides = {
        get_spawner_manager: lambda: spawner_manager,
        get_bridge_name: lambda: "mock",
        get_bridge_manager: lambda: bridge_manager,
        get_spawner_name: lambda: "mock",
    }
    yield APP

    result = runner.invoke(drop, ["--yes"])
    assert result.exit_code == 0


@pytest.fixture
def client(app):
    return TestClient(app)



from moriarty.matrix.operator_.params import (
    CreateEndpointParams,
    SetAutoscaleParams,
    UpdateEndpointParams,
)


def test_hello(client):
    response = client.get("/")
    assert response.status_code == 200


def test_manage_endpoints(client):
    endpoint_name = "hello"

    response = client.get("/endpoint/list")
    assert response.status_code == 200
    assert len(response.json()["endpoints"]) == 0

    response = client.get(f"/endpoint/{endpoint_name}/info")
    assert response.status_code == 404

    response = client.post(
        "/endpoint/create",
        data=CreateEndpointParams(endpoint_name=endpoint_name, image="busybox").model_dump_json(),
    )
    assert response.status_code == 201

    response = client.post(
        "/endpoint/create",
        data=CreateEndpointParams(endpoint_name=endpoint_name, image="busybox").model_dump_json(),
    )
    assert response.status_code == 409

    response = client.get(f"/endpoint/{endpoint_name}/info")
    assert response.status_code == 200

    response = client.get("/endpoint/list")
    assert response.status_code == 200
    assert len(response.json()["endpoints"]) == 1

    response = client.get("/endpoint/list", params={"cursor": endpoint_name})
    assert response.status_code == 200
    assert len(response.json()["endpoints"]) == 0

    response = client.get("/endpoint/list", params={"keyword": endpoint_name})
    assert response.status_code == 200
    assert len(response.json()["endpoints"]) == 1

    response = client.get("/endpoint/list", params={"keyword": "not-exist"})
    assert response.status_code == 200
    assert len(response.json()["endpoints"]) == 0

    for order_by in ["created_at", "updated_at", "endpoint_name"]:
        for order in ["asc", "desc"]:
            response = client.get(
                "/endpoint/list",
                params={"orderBy": order_by, "order": order},
            )
            assert response.status_code == 200

    response = client.post(
        f"/endpoint/{endpoint_name}/update",
        data=UpdateEndpointParams(image="busybox", queue_capacity=10).model_dump_json(),
    )
    assert response.status_code == 200

    response = client.post(
        "/endpoint/not-exist/update",
        data=UpdateEndpointParams(image="busybox", queue_capacity=10).model_dump_json(),
    )
    assert response.status_code == 404

    response = client.post(f"/endpoint/{endpoint_name}/delete")
    assert response.status_code == 204
    response = client.get(f"/endpoint/{endpoint_name}/info")
    assert response.status_code == 404


def test_manage_autoscale(client):
    endpoint_name = "hello-scaleable"

    response = client.post(
        "/endpoint/create",
        data=CreateEndpointParams(endpoint_name=endpoint_name, image="busybox").model_dump_json(),
    )
    assert response.status_code == 201

    response = client.get(f"/autoscale/{endpoint_name}/info")
    assert response.status_code == 404

    response = client.post(
        f"/autoscale/{endpoint_name}/set", data=SetAutoscaleParams().model_dump_json()
    )
    assert response.status_code == 200
    response = client.get(f"/autoscale/{endpoint_name}/info")
    assert response.status_code == 200

    response = client.post(f"/autoscale/{endpoint_name}/delete")
    assert response.status_code == 204

    response = client.get(f"/autoscale/{endpoint_name}/info")
    assert response.status_code == 404



from kubernetes_asyncio import config

from moriarty.log import logger

_config_loaded = False


async def load_kube_config() -> None:
    global _config_loaded
    if _config_loaded:
        return
    try:
        logger.debug("Try to load in-cluster config")
        config.load_incluster_config()
    except config.config_exception.ConfigException:
        logger.info("No in-cluster config, using default config location")
        await config.load_kube_config()
    _config_loaded = True



import os
from datetime import datetime
from urllib.parse import urljoin

import httpx

from moriarty.envs import MORIARTY_CONNECTOR_ENDPOINT_ENV
from moriarty.log import logger
from moriarty.matrix.connector.params import InvokeParams


async def invoke(
    endpoint_name: str,
    invoke_params: dict,
    inference_id: str = None,
    priority: int = None,
    created_at: datetime | int = None,
    expires_at: datetime | int = None,
    endpoint: str = os.getenv(MORIARTY_CONNECTOR_ENDPOINT_ENV, "http://moriarty-connector:8901"),
    *,
    async_client=httpx.AsyncClient,
    **client_kwargs,
) -> httpx.Response:
    """
    Invoke an inference endpoint.

    Response: httpx.Response
    """
    url = urljoin(endpoint, "/invoke")
    if expires_at and isinstance(expires_at, int):
        expires_at = datetime.fromtimestamp(expires_at)

    if created_at and isinstance(created_at, int):
        created_at = datetime.fromtimestamp(created_at)

    payload = InvokeParams(
        inference_id=inference_id,
        endpoint_name=endpoint_name,
        invoke_params=invoke_params,
        priority=priority,
        created_at=created_at or datetime.now(),
        expires_at=expires_at,
    )
    async with async_client(
        follow_redirects=True,
    ) as client:
        logger.debug(f"Invoke endpoint: {url} with payload: {payload}")
        return await client.post(
            url,
            data=payload.model_dump_json(),
            timeout=60,
            **client_kwargs,
        )



import asyncio
from functools import wraps

import click
import uvicorn

from moriarty.matrix.operator_.config import get_config
from moriarty.matrix.operator_.dbutils import (
    drop_all_data,
    get_db_url,
    upgrade_in_place,
)

from .api_app import app as api_app
from .callback_app import app as callback_app
from .daemon import BridgeDaemon, KubeAutoscalerDaemon


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.command()
@click.option("--host", type=click.STRING, default="0.0.0.0")
@click.option("--port", type=click.INT, default=8999)
def callback(host, port):
    """
    Start the server.
    """
    uvicorn.run(callback_app, host=host, port=port)


@click.command()
@click.option("--host", type=click.STRING, default="0.0.0.0")
@click.option("--port", type=click.INT, default=8902)
def api(host, port):
    """
    Start the server.
    """
    uvicorn.run(api_app, host=host, port=port)


@click.command()
@coro
async def autoscale():
    """
    Start autoscaling daemon for k8s.
    """
    await KubeAutoscalerDaemon(get_config()).run_forever()


@click.command()
@coro
async def bridge():
    """
    Start bridge daemon.
    """
    await BridgeDaemon(get_config()).run_forever()


@click.command()
def init():
    """
    Init and upgrade the database.
    """

    config = get_config()

    upgrade_in_place(
        db_url=get_db_url(config, async_mode=False),
    )


@click.command()
@click.option("--yes", "-y", is_flag=True, default=False)
def drop(yes):
    """
    Drop all data before testing or other purposes.

    This command is not visible in the CLI. Only use it in tests for now.
    """
    if not yes:
        click.confirm("Are you sure you want to drop all data?", abort=True)

    click.echo("Dropping all data...")

    config = get_config()
    db_url = get_db_url(config, async_mode=False)
    drop_all_data(db_url)


@click.group()
def cli():
    pass


cli.add_command(init)
cli.add_command(bridge)
cli.add_command(api)
cli.add_command(autoscale)
cli.add_command(callback)
# cli.add_command(drop) # noqa: not visible in CLI



from __future__ import annotations

import os
from typing import Any, Callable

from pydantic import BaseModel

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

from moriarty.envs import *


def get_config() -> Config:
    default = {
        DB_HOST_ENV: "localhost",
        DB_PORT_ENV: 5432,
        DB_USER_ENV: "postgres",
        DB_PASSWORD_ENV: "Th3PAssW0rd!1.",
        DB_DATABASE_ENV: "moriarty-matrix",
        REDIS_HOST_ENV: "localhost",
        REDIS_PORT_ENV: 6379,
        REDIS_DB_ENV: 0,
        REDIS_TLS_ENV: False,
    }

    return Config.from_env(default)


class DBConfig(BaseModel):
    host: str
    port: int
    user: str
    password: str
    database: str

    @classmethod
    def from_env(cls, default: dict[str, str]) -> "DBConfig":
        host = os.getenv(DB_HOST_ENV, default.get(DB_HOST_ENV, "localhost"))
        port = int(os.getenv(DB_PORT_ENV, default.get(DB_PORT_ENV, 5432)))
        user = os.getenv(DB_USER_ENV, default.get(DB_USER_ENV, "postgres"))
        password = os.getenv(DB_PASSWORD_ENV, default.get(DB_PASSWORD_ENV, "postgres"))
        database = os.getenv(DB_DATABASE_ENV, default.get(DB_DATABASE_ENV, "postgres"))
        return cls(
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
        )

    def get_db_url(self):
        return f"{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"

    def get_log_db_url(self):
        return self.get_db_url().replace(self.password, "********")


class RedisConfig(BaseModel):
    host: str
    port: int
    password: str
    db: int
    tls: bool
    cluster: bool

    @classmethod
    def from_env(cls, default: dict[str, str]) -> "RedisConfig":
        host = os.getenv(REDIS_HOST_ENV, default.get(REDIS_HOST_ENV, "localhost"))
        port = int(os.getenv(REDIS_PORT_ENV, default.get(REDIS_PORT_ENV, 6379)))
        password = os.getenv(REDIS_PASSWORD_ENV, default.get(REDIS_PASSWORD_ENV, ""))
        db = int(os.getenv(REDIS_DB_ENV, default.get(REDIS_DB_ENV, 6)))
        tls = (
            os.getenv(REDIS_TLS_ENV) in ["True", "true"]
            if os.getenv(REDIS_TLS_ENV)
            else default.get(REDIS_TLS_ENV, False)
        )
        cluster = (
            os.getenv(REDIS_CLUSTER_ENV) in ["True", "true"]
            if os.getenv(REDIS_CLUSTER_ENV)
            else default.get(REDIS_CLUSTER_ENV, False)
        )
        return cls(
            host=host,
            port=port,
            password=password,
            db=db,
            tls=tls,
            cluster=cluster,
        )

    def get_redis_url(self):
        if not self.cluster:
            url = f"{self.host}:{self.port}/{self.db}"
        else:
            # Ignore db
            url = f"{self.host}:{self.port}"

        if self.tls:
            return f"rediss://{url}"
        else:
            return f"redis://{url}"


class Config(BaseModel):
    db: DBConfig
    redis: RedisConfig

    """
    Cache for latest initialized object.
    """

    @classmethod
    def from_env(cls, default: None | dict[str, Any]) -> "Config":
        if default is None:
            default = {}
        return cls(
            db=DBConfig.from_env(default=default),
            redis=RedisConfig.from_env(default=default),
        )



import click

from moriarty.envs import MORIARTY_MATRIX_API_URL_ENV, MORIARTY_MATRIX_TOKEN_ENV
from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.matrix.operator_.params import CreateEndpointParams, UpdateEndpointParams

from .sdk import (
    request_create_endpoint_with_params,
    request_delete_autoscale,
    request_delete_endpoint,
    request_query_autoscale,
    request_query_endpoint,
    request_scan_autoscale_log,
    request_scan_endpoint,
    request_set_autoscale,
    request_update_endpoint_with_params,
)


@click.command()
@click.option(
    "--limit",
    default=10,
    help="Limit number of results",
    type=int,
)
@click.option(
    "--cursor",
    default=None,
    help="Cursor",
    type=str,
)
@click.option(
    "--keyword",
    default=None,
    help="Keyword",
    type=str,
)
@click.option(
    "--order_by",
    default="created_at",
    help="Order by",
    type=str,
)
@click.option(
    "--order",
    default="desc",
    help="Order",
    type=str,
)
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def scan(
    api_url,
    token,
    limit,
    cursor,
    keyword,
    order_by,
    order,
):
    cursor = cursor
    while True:
        response = request_scan_endpoint(
            api_url=api_url,
            limit=limit,
            cursor=cursor,
            keyword=keyword,
            order_by=order_by,
            order=order,
            token=token,
        )

        print(
            f"Count: {len(response.endpoints)} | Total: {response.total} [Cursor: {response.cursor}]"
        )
        cursor = response.cursor

        for endpoint in response.endpoints:
            autoscale = request_query_autoscale(
                endpoint_name=endpoint.endpoint_name, api_url=api_url, token=token
            )
            print(f"Endpoint: {endpoint} | Autoscale: {autoscale}")

        if not response.endpoints:
            print("\nNo more results...")
            return
        input("Press Enter to continue...")


@click.command()
@click.argument("endpoint_name")
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def query(
    endpoint_name,
    api_url,
    token,
):
    response = request_query_endpoint(
        endpoint_name=endpoint_name,
        api_url=api_url,
        token=token,
    )
    print(response)

    autoscale = request_query_autoscale(endpoint_name=endpoint_name, api_url=api_url, token=token)
    if autoscale:
        print(autoscale)
    else:
        print(f"No autoscale found for `{endpoint_name}`")


@click.command()
@click.option(
    "--endpoint-config-file",
    help="Endpoint config file",
    type=str,
    required=True,
)
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def deploy(
    api_url,
    endpoint_config_file,
    token,
):
    with open(endpoint_config_file) as f:
        params: CreateEndpointParams = CreateEndpointParams.model_validate_json(
            f.read(),
        )

    request_create_endpoint_with_params(api_url=api_url, params=params, token=token)
    print(f"Created endpoint: {params.endpoint_name}")


@click.command()
@click.argument("endpoint_name")
@click.option(
    "--endpoint-config-file",
    help="Endpoint config file",
    type=str,
    required=True,
)
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def update(
    api_url,
    endpoint_name,
    endpoint_config_file,
    token,
):
    with open(endpoint_config_file) as f:
        params: UpdateEndpointParams = UpdateEndpointParams.model_validate_json(
            f.read(),
        )

    autoscale = request_query_autoscale(endpoint_name=endpoint_name, api_url=api_url, token=token)
    if autoscale:
        print(f"`{endpoint_name}` has autoscale: {autoscale}, set replicas to None")
        params.replicas = None

    response = request_update_endpoint_with_params(
        api_url=api_url,
        endpoint_name=endpoint_name,
        params=params,
        token=token,
    )
    print(f"Updated endpoint: `{endpoint_name}`")
    print(response)


@click.command()
@click.argument("endpoint_name")
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def delete(
    endpoint_name,
    api_url,
    token,
):
    request_delete_endpoint(endpoint_name=endpoint_name, api_url=api_url, token=token)
    print(f"Endpoint `{endpoint_name}` deleted")


@click.command()
@click.argument("endpoint_name")
@click.option(
    "--min-replicas",
    required=True,
    help="Minimum number of replicas",
    type=int,
)
@click.option(
    "--max-replicas",
    required=True,
    help="Minimum number of replicas",
    type=int,
)
@click.option(
    "--scale-in-cooldown",
    required=True,
    help="Scale in cooldown",
    type=int,
)
@click.option(
    "--scale-out-cooldown",
    required=True,
    help="Scale out cooldown",
    type=int,
)
@click.option(
    "--metrics",
    required=True,
    default=MetricType.pending_jobs_per_instance.value,
    help="Metrics",
    type=click.Choice(MetricType.__members__.keys()),
)
@click.option(
    "--metrics-threshold",
    required=True,
    help="Metrics threshold",
    type=int,
)
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def autoscale(
    endpoint_name,
    min_replicas,
    max_replicas,
    scale_in_cooldown,
    scale_out_cooldown,
    metrics,
    metrics_threshold,
    api_url,
    token,
):
    request_set_autoscale(
        endpoint_name=endpoint_name,
        min_replicas=min_replicas,
        max_replicas=max_replicas,
        scale_in_cooldown=scale_in_cooldown,
        scale_out_cooldown=scale_out_cooldown,
        metrics=metrics,
        metrics_threshold=metrics_threshold,
        api_url=api_url,
        token=token,
    )

    autoscale = request_query_autoscale(endpoint_name=endpoint_name, api_url=api_url, token=token)
    if autoscale:
        print("Autoscale set successfully")
        print(autoscale)
    else:
        print(f"No autoscale found for `{endpoint_name}`")


@click.command()
@click.argument("endpoint_name")
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def delete_autoscale(
    endpoint_name,
    api_url,
    token,
):
    request_query_endpoint(
        endpoint_name=endpoint_name,
        api_url=api_url,
        token=token,
    )

    request_delete_autoscale(endpoint_name=endpoint_name, api_url=api_url, token=token)
    print(f"Autoscale for `{endpoint_name}` deleted")
    response = request_query_endpoint(
        endpoint_name=endpoint_name,
        api_url=api_url,
        token=token,
    )
    print(response)


@click.command()
@click.argument("endpoint_name")
@click.option(
    "--limit",
    default=10,
    help="Limit number of results",
    type=int,
)
@click.option(
    "--cursor",
    default=None,
    help="Cursor",
    type=int,
)
@click.option(
    "--keyword",
    default=None,
    help="Keyword",
    type=str,
)
@click.option(
    "--order_by",
    default="created_at",
    help="Order by",
    type=str,
)
@click.option(
    "--order",
    default="desc",
    help="Order",
    type=str,
)
@click.option(
    "--api-url",
    help="Moriarty Operator API URL",
    type=str,
    envvar=MORIARTY_MATRIX_API_URL_ENV,
    required=True,
)
@click.option(
    "--token",
    help="Moriarty Operator API token",
    type=str,
    envvar=MORIARTY_MATRIX_TOKEN_ENV,
)
def query_autoscale_log(
    endpoint_name,
    api_url,
    token,
    limit,
    cursor,
    keyword,
    order_by,
    order,
):
    cursor = cursor
    while True:
        response = request_scan_autoscale_log(
            api_url=api_url,
            endpoint_name=endpoint_name,
            limit=limit,
            cursor=cursor,
            keyword=keyword,
            order_by=order_by,
            order=order,
            token=token,
        )

        print(f"Count: {len(response.logs)} | Total: {response.total} [Cursor: {response.cursor}]")
        cursor = response.cursor

        for log in response.logs:
            print(log)

        if not response.logs:
            print("\nNo more results...")
            return
        input("Press Enter to continue...")


@click.group()
def cli():
    pass


cli.add_command(scan)
cli.add_command(query)
cli.add_command(deploy)
cli.add_command(update)
cli.add_command(delete)
cli.add_command(autoscale)
cli.add_command(delete_autoscale)
cli.add_command(query_autoscale_log)



from fastapi import FastAPI, Request, Response, status

app = FastAPI()


@app.get("/ping")
async def root():
    """Leave it for health check"""
    return Response(status_code=status.HTTP_200_OK)


@app.post("/invocations")
async def root(
    request: Request,
):
    return await request.json()



import os

MORIARTY_BRIDGE_NAME_ENV = "MORIARTY_BRIDGE_NAME"
MORIARTY_SPAWNER_NAME_ENV = "MORIARTY_SPAWNER_NAME"
MORIARTY_MATRIX_TOKEN_ENV = "MORIARTY_MATRIX_TOKEN"
MORIARTY_CONNECTOR_ENDPOINT_ENV = "MORIARTY_CONNECTOR_ENDPOINT"
MORIARTY_BRIDGE_RESULT_QUEUE_URL_ENV = "MORIARTY_BRIDGE_RESULT_QUEUE_URL"

MORIARTY_KUBE_NAMESPACE_ENV = "MORIARTY_KUBE_NAMESPACE"


def get_bridge_name() -> str:
    return os.getenv(MORIARTY_BRIDGE_NAME_ENV, "sqs")


def get_bridge_result_queue_url() -> str:
    return os.getenv(MORIARTY_BRIDGE_RESULT_QUEUE_URL_ENV, "")


def get_spawner_name() -> str:
    return os.getenv(MORIARTY_SPAWNER_NAME_ENV, "kube")


DB_HOST_ENV = "DB_HOST"
DB_PORT_ENV = "DB_PORT"
DB_USER_ENV = "DB_USER"
DB_PASSWORD_ENV = "DB_PASSWORD"
DB_DATABASE_ENV = "DB_DATABASE"


REDIS_URL_ENV = "REDIS_URL"
REDIS_HOST_ENV = "REDIS_HOST"
REDIS_PORT_ENV = "REDIS_PORT"
REDIS_DB_ENV = "REDIS_DB"
REDIS_CLUSTER_ENV = "REDIS_CLUSTER"
REDIS_TLS_ENV = "REDIS_TLS"
REDIS_USERNAME_ENV = "REDIS_USERNAME"
REDIS_PASSWORD_ENV = "REDIS_PASSWORD"


ENDPOINT_NAME_ENV = "ENDPOINT_NAME"
INVOKE_HOST_ENV = "INVOKE_HOST"
INVOKE_PORT_ENV = "INVOKE_PORT"
INVOKE_PATH_ENV = "INVOKE_PATH"
HEALTH_CHECK_PATH_ENV = "HEALTH_CHECK_PATH"
ENABLE_RETRY_ENV = "ENABLE_RETRY"
ENABLE_SSL_ENV = "ENABLE_SSL"
CONCURRENCY_ENV = "CONCURRENCY"

PROCESS_TIMEOUT_ENV = "PROCESS_TIMEOUT"
HEALTH_CHECK_TIMEOUT_ENV = "HEALTH_CHECK_TIMEOUT"
HEALTH_CHECK_INTERVAL_ENV = "HEALTH_CHECK_INTERVAL"


MODEL_S3_ACCESS_KEY_ID_ENV = "MODEL_S3_ACCESS_KEY_ID"
MODEL_S3_SECRET_ACCESS_KEY_ENV = "MODEL_S3_SECRET_ACCESS_KEY"
MODEL_S3_ENDPOINT_URL_ENV = "MODEL_S3_ENDPOINT_URL"
MODEL_AWS_REGION_ENV = "MODEL_AWS_REGION"

MORIARTY_MATRIX_API_URL_ENV = "MORIARTY_MATRIX_API_URL"



from __future__ import annotations

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

import glob
import importlib
from os.path import basename, dirname, isfile, join
from typing import Any

import pluggy

from moriarty.exceptions import PluginInitializationError, PluginNotFoundError
from moriarty.log import logger
from moriarty.matrix.job_manager.bridge import impl, plugin


class BridgeManager:
    def __init__(self):
        self.pm = pluggy.PluginManager(plugin.project_name)
        self.pm.add_hookspecs(plugin)
        self._registed_cls: dict[str, type[plugin.QueueBridge]] = {}

        self.pm.load_setuptools_entrypoints(plugin.project_name)
        self.load_all_local_model()

    def load_all_local_model(self):
        """
        Import all implementation in ./impl
        """
        logger.debug(f"Loading all local model from {impl}")

        return self._load_dir(impl)

    @property
    def registed_cls(self) -> dict[str, type]:
        """
        Access all registed class.

        Lazy load, only load once.
        """
        if self._registed_cls:
            return self._registed_cls
        for f in self.pm.hook.register(manager=self):
            try:
                f()
            except Exception as e:
                logger.error(f"One bridge registration failed: {e}")
                logger.exception(e)
                continue
        return self._registed_cls

    def _load_dir(self, module):
        """
        Import all python files in a submodule.
        """
        modules = glob.glob(join(dirname(module.__file__), "*.py"))
        sub_packages = (
            basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
        )
        packages = (str(module.__package__) + "." + i for i in sub_packages)
        for p in packages:
            logger.debug(f"Loading {p}")
            self.pm.register(importlib.import_module(p))

    def _normalize_name(self, name: str) -> str:
        return name.strip().lower()

    def register(self, cls: type[plugin.QueueBridge]):
        """
        Register a new module, if the module is already registed, skip it.
        """

        cls_name = self._normalize_name(cls.register_name)
        logger.debug(f"Register for new module: {cls_name}")
        if cls in self._registed_cls.values():
            logger.warning(
                f"SKIP: trying to register {cls} as '{cls_name}', but {cls_name} is already registed."
            )
            return
        if not issubclass(cls, plugin.QueueBridge):
            logger.error(f"SKIP: {cls_name} is not a subclass of {plugin.QueueBridge}")
            return
        self._registed_cls[cls_name] = cls

    @cache
    def init(self, c, *args, **kwargs: dict[str, Any]):
        """
        Init a new subclass of plugin.QueueBridge.

        Raises:
            NotFoundError: if cls_name is not registered
            InitializationError: if failed to initialize
        """
        if isinstance(c, plugin.QueueBridge):
            return c

        if isinstance(c, type):
            cls_type = c
        else:
            c = self._normalize_name(c)

            if not c in self.registed_cls:
                raise PluginNotFoundError(f"{c} is not registed.")
            cls_type = self.registed_cls[c]
        try:
            return cls_type(*args, **kwargs)
        except Exception as e:
            raise PluginInitializationError(e)



from __future__ import annotations

import socket
from functools import partial
from typing import TYPE_CHECKING

import httpx
import pytest
from fastapi.testclient import TestClient

import docker
from moriarty import mock
from moriarty.envs import get_bridge_name
from moriarty.matrix.connector.app import app as APP
from moriarty.matrix.job_manager.bridge.manager import BridgeManager
from moriarty.matrix.job_manager.bridge_wrapper import get_bridge_manager

if TYPE_CHECKING:
    from docker import DockerClient


def get_port():
    # Get an unoccupied port
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def docker_client():
    try:
        client = docker.from_env()
        client.ping()
        return client
    except:
        pytest.skip("Docker is not available")


@pytest.fixture
def bridge_manager():
    bm = BridgeManager()
    bm._load_dir(mock)
    yield bm


@pytest.fixture
async def app(bridge_manager):
    APP.dependency_overrides = {
        get_bridge_name: lambda: "mock",
        get_bridge_manager: lambda: bridge_manager,
    }
    yield APP


@pytest.fixture
def client(app):
    return TestClient(app)


@pytest.fixture
async def async_client(app):
    return partial(httpx.AsyncClient, app=app)



from moriarty.matrix.connector.params import InvokeParams
from moriarty.matrix.connector.sdk import invoke


def test_hello(client):
    response = client.get("/")
    assert response.status_code == 200


def test_invoke(client):
    response = client.post(
        "invoke",
        data=InvokeParams(
            inference_id="test",
            endpoint_name="hello",
            invoke_params={"name": "World"},
            priority=9,
        ).model_dump_json(),
    )
    assert response.status_code == 200
    assert response.json() == {"inference_id": "test"}


async def test_sdk_invoke(async_client):
    response = await invoke(
        endpoint_name="hello",
        inference_id="test",
        invoke_params={"name": "World"},
        async_client=async_client,
    )
    assert response.status_code == 200
    assert response.json() == {"inference_id": "test"}



from __future__ import annotations

import math
from functools import cached_property

import redis.asyncio as redis
from pydantic import BaseModel
from sqlalchemy import delete, func, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.matrix.operator_.orm import AutoscalerORM, EndpointORM
from moriarty.matrix.operator_.spawner import plugin
from moriarty.sidecar.producer import JobProducer


class EndpointMetrics(BaseModel):
    endpoint_name: str
    metrics: dict[MetricType, float] = dict()


class MetricsManager:
    def __init__(self, spawner: plugin.Spawner, redis_client: redis.Redis) -> None:
        self.spawner = spawner
        self.redis_client = redis_client

    @cached_property
    def job_producer(self) -> JobProducer:
        return JobProducer(redis_client=self.redis_client)

    async def count_pending_jobs(self, endpoint_name: str) -> int:
        return await self.job_producer.count_unprocessed_jobs(endpoint_name=endpoint_name)

    async def get_metrics(self, endpoint_name: str) -> EndpointMetrics:
        unprocessed_count = await self.count_pending_jobs(endpoint_name)
        replicas = max(1, await self.spawner.count_avaliable_instances(endpoint_name))
        return EndpointMetrics(
            endpoint_name=endpoint_name,
            metrics={
                MetricType.pending_jobs: unprocessed_count,
                MetricType.pending_jobs_per_instance: unprocessed_count / replicas,
            },
        )

    def calculate_least_replicas(
        self,
        metric: MetricType,
        current_metric_value: float | int,
        metric_threshold: float,
    ) -> int:
        if metric == MetricType.pending_jobs:
            return int(current_metric_value - int(metric_threshold))
        if metric == MetricType.pending_jobs_per_instance:
            if metric_threshold == 0:
                metric_threshold = 1
            return math.ceil(current_metric_value / metric_threshold)

        raise NotImplementedError

    def calculate_queue_capacity(self, autoscaler: AutoscalerORM) -> int:
        metric: MetricType = autoscaler.metric
        if metric == MetricType.pending_jobs:
            return autoscaler.metrics_threshold + autoscaler.max_replicas + 1
        if metric == MetricType.pending_jobs_per_instance:
            return math.ceil(autoscaler.metrics_threshold) * autoscaler.max_replicas + 1

        raise NotImplementedError


class EndpointMixin:
    session: AsyncSession

    async def get_endpoint_orm(self, endpoint_name: str) -> EndpointORM | None:
        return (
            await self.session.execute(
                select(EndpointORM).where(EndpointORM.endpoint_name == endpoint_name)
            )
        ).scalar_one_or_none()

    async def get_avaliable_endpoints(self) -> list[str]:
        endpoint_names = (
            (
                await self.session.execute(
                    select(EndpointORM.endpoint_name).where(
                        EndpointORM.replicas > 0,
                    )
                )
            )
            .scalars()
            .all()
        )
        return endpoint_names


class AutoscaleMixin:
    async def get_autoscaler_orm(self, endpoint_name: str) -> AutoscalerORM | None:
        return (
            await self.session.execute(
                select(AutoscalerORM).where(AutoscalerORM.endpoint_name == endpoint_name)
            )
        ).scalar_one_or_none()



from moriarty.matrix.operator_.spawner.plugin import Spawner, hookimpl


class CustomSpawner(Spawner):
    register_name = "example"


@hookimpl
def register(manager):
    manager.register(CustomSpawner)



from custom_spawner.imp import CustomSpawner

from moriarty.matrix.operator_.spawner.manager import SpawnerManager


def test_registed():
    manager = SpawnerManager()

    assert CustomSpawner.register_name in manager.registed_cls.keys()



import threading


class Singleton(type):
    """
    metaclass for singleton, thread-safe.
    """

    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



import os
from contextlib import asynccontextmanager

from fastapi import Depends, FastAPI, HTTPException, Query, Response

from moriarty.envs import MORIARTY_MATRIX_TOKEN_ENV
from moriarty.log import logger
from moriarty.matrix.connector.invoker import Invoker, get_invoker
from moriarty.matrix.connector.params import InvokeParams, InvokeResponse

TOKEN = os.getenv(MORIARTY_MATRIX_TOKEN_ENV)


@asynccontextmanager
async def lifespan(app: FastAPI):
    if TOKEN:
        logger.info(f"Token: {TOKEN}")
    else:
        logger.info("Token not set, authentication is disabled")

    yield


app = FastAPI(lifespan=lifespan)


@app.middleware("http")
async def verify_token(request, call_next):
    if request.url.path == "/":
        return await call_next(request)

    if TOKEN and request.headers.get("Authorization") != f"Bearer {TOKEN}":
        return Response(
            status_code=401,
            content=f"Unauthorized. Check environment {MORIARTY_MATRIX_TOKEN_ENV} for token.",
        )
    return await call_next(request)


@app.post("/invoke")
async def invoke(
    params: InvokeParams,
    invoker: Invoker = Depends(get_invoker),
) -> InvokeResponse:
    return await invoker.invoke(params)


@app.get("/")
async def root():
    return {"message": "Hello World"}



from moriarty.log import logger
from moriarty.matrix.operator_.orm import EndpointORM
from moriarty.matrix.operator_.params import EndpointRuntimeInfo
from moriarty.matrix.operator_.spawner.plugin import Spawner, hookimpl


class MockSpawner(Spawner):
    register_name = "mock"

    async def prepare(self) -> None:
        pass

    async def count_avaliable_instances(self, endpoint_name: str) -> int:
        return 1

    async def create(self, endpoint_orm: EndpointORM) -> None:
        return None

    async def update(self, endpoint_orm: EndpointORM, need_restart: bool = False) -> None:
        return None

    async def delete(self, endpoint_name: str) -> None:
        return None

    async def get_runtime_info(self, endpoint_name: str) -> EndpointRuntimeInfo:
        return EndpointRuntimeInfo(
            total_replicas_nums=1,
            updated_replicas_nums=1,
            avaliable_replicas_nums=1,
            unavailable_replicas_nums=0,
        )

    async def scale(self, endpoint_name: str, target_replicas: int) -> None:
        return None


@hookimpl
def register(manager):
    manager.register(MockSpawner)



from moriarty.matrix.job_manager.bridge.manager import BridgeManager


def test_bridge_manager():
    registed = ["sqs"]
    manager = BridgeManager()

    assert sorted(registed) == sorted(manager.registed_cls.keys())

    sqs_bridge = manager.init("sqs")
    assert sqs_bridge == manager.init("sqs")



import pytest

from moriarty.matrix.job_manager.bridge.impl.sqs_bridge import SQSBridge
from moriarty.matrix.job_manager.params import (
    InferenceJob,
    InferenceResult,
    InferenceResultStatus,
)


@pytest.fixture
def endpoint_name(case_id):
    return f"test-{case_id}"


@pytest.fixture
def priority():
    return 1


@pytest.fixture
def sqs_bridge(case_id, endpoint_name, priority):
    try:
        import boto3

        sqs = boto3.client("sqs")
    except ImportError:
        pytest.skip("boto3 is not installed")
    except Exception:
        pytest.skip("boto3 is not configured")

    job_queue_name = f"moriarty-bridge-{endpoint_name}-{priority}"
    result_queue_name = f"moriarty-bridge-result-{case_id}"
    try:
        job_queue_url = sqs.create_queue(QueueName=job_queue_name)["QueueUrl"]
        result_queue_url = sqs.create_queue(QueueName=result_queue_name)["QueueUrl"]
    except sqs.exceptions.QueueAlreadyExists:
        pass
    except Exception:
        pytest.skip("boto3 is not configured")

    bridge = SQSBridge(
        bridge_result_queue_url=result_queue_url,
    )
    bridge.make_job_queue_url = lambda endpoint_name, priority: job_queue_url
    yield bridge
    try:
        sqs.delete_queue(QueueUrl=job_queue_url)
        sqs.delete_queue(QueueUrl=result_queue_url)
    except sqs.exceptions.QueueDoesNotExist:
        pass


async def test_bridge_job(sqs_bridge: SQSBridge):
    _called = False
    input_job = InferenceJob(
        inference_id="test",
        payload={"input": "test"},
    )
    job = await sqs_bridge.enqueue_job("test", job=input_job)

    assert job

    async def _process_func(job: InferenceJob):
        nonlocal _called
        _called = True
        assert job == input_job

    await sqs_bridge.dequeue_job("test", _process_func)

    assert _called


@pytest.mark.parametrize(
    "inference_result",
    [
        InferenceResult(
            inference_id="test",
            status=InferenceResultStatus.COMPLETED,
            message="Oh yeah",
            payload='{"output": "test"}',
        ),
        InferenceResult(
            inference_id="test",
            status=InferenceResultStatus.FAILED,
            message="Oh no",
            payload='{"error": "test"}',
        ),
    ],
)
async def test_bridge_result(
    sqs_bridge: SQSBridge,
    inference_result: InferenceResult,
    endpoint_name,
    priority,
):
    _called = False

    await sqs_bridge.enqueue_result(inference_result)

    async def _process_func(result: InferenceResult):
        nonlocal _called
        _called = True
        assert result == inference_result

    await sqs_bridge.list_avaliable_priorities(endpoint_name) == [priority]
    await sqs_bridge.dequeue_result(_process_func)

    assert _called

    await sqs_bridge.list_avaliable_priorities(endpoint_name) == []



import click
from brq.tools import get_redis_client, get_redis_url

from moriarty.envs import *
from moriarty.sidecar.consumer import InferencerConsumer
from moriarty.tools import coro


@click.command()
@click.option(
    "--redis-host",
    default="localhost",
    help="Redis host",
    type=str,
    envvar=REDIS_HOST_ENV,
)
@click.option(
    "--redis-port",
    default=6379,
    help="Redis port",
    type=int,
    envvar=REDIS_PORT_ENV,
)
@click.option(
    "--redis-db",
    default=0,
    help="Redis db",
    type=int,
    envvar=REDIS_DB_ENV,
)
@click.option(
    "--redis-cluster",
    default=False,
    help="Redis cluster",
    type=bool,
    envvar=REDIS_CLUSTER_ENV,
)
@click.option(
    "--redis-tls",
    default=False,
    help="Redis TLS",
    type=bool,
    envvar=REDIS_TLS_ENV,
)
@click.option(
    "--redis-username",
    default="",
    help="Redis username",
    type=str,
    envvar=REDIS_USERNAME_ENV,
)
@click.option(
    "--redis-password",
    default="",
    help="Redis password",
    type=str,
    envvar=REDIS_PASSWORD_ENV,
)
@click.option("--redis-url", required=False, help="Redis URL", type=str, envvar=REDIS_URL_ENV)
@click.option(
    "--endpoint-name",
    required=True,
    help="Endpoint name",
    type=str,
    envvar=ENDPOINT_NAME_ENV,
)
@click.option(
    "--invoke-host",
    required=False,
    help="Invoke host",
    type=str,
    envvar=INVOKE_HOST_ENV,
)
@click.option(
    "--invoke-port",
    required=False,
    help="Invoke port",
    type=int,
    envvar=INVOKE_PORT_ENV,
)
@click.option(
    "--invoke-path",
    required=False,
    help="Invoke path",
    type=str,
    envvar=INVOKE_PATH_ENV,
)
@click.option(
    "--health-check-path",
    required=False,
    help="Health check url",
    type=str,
    envvar=HEALTH_CHECK_PATH_ENV,
)
@click.option(
    "--enable-retry",
    required=False,
    help="Enable retry",
    type=bool,
    envvar=ENABLE_RETRY_ENV,
)
@click.option(
    "--enable-ssl",
    required=False,
    help="Enable SSL",
    type=bool,
    envvar=ENABLE_SSL_ENV,
)
@click.option(
    "--concurrency",
    required=False,
    help="Concurrency",
    type=int,
    envvar=CONCURRENCY_ENV,
)
@click.option(
    "--process-timeout",
    required=False,
    help="Process timeout",
    type=int,
    envvar=PROCESS_TIMEOUT_ENV,
)
@click.option(
    "--health-check-timeout",
    required=False,
    help="Health check timeout",
    type=int,
    envvar=HEALTH_CHECK_TIMEOUT_ENV,
)
@click.option(
    "--health-check-interval",
    required=False,
    help="Health check interval",
    type=int,
    envvar=HEALTH_CHECK_INTERVAL_ENV,
)
@coro
async def start(
    redis_host,
    redis_port,
    redis_db,
    redis_cluster,
    redis_tls,
    redis_username,
    redis_password,
    redis_url,
    endpoint_name,
    invoke_host,
    invoke_port,
    invoke_path,
    health_check_path,
    enable_retry,
    enable_ssl,
    concurrency,
    process_timeout,
    health_check_timeout,
    health_check_interval,
):
    """
    Start moriarty sidecar and run forever
    """
    redis_url = redis_url or get_redis_url(
        host=redis_host,
        port=redis_port,
        db=redis_db,
        cluster=redis_cluster,
        tls=redis_tls,
        username=redis_username,
        password=redis_password,
    )
    kwargs = {
        k: v
        for k, v in {
            "invoke_host": invoke_host,
            "invoke_port": invoke_port,
            "invoke_path": invoke_path,
            "health_check_path": health_check_path,
            "enable_retry": enable_retry,
            "enable_ssl": enable_ssl,
            "concurrency": concurrency,
            "process_timeout": process_timeout,
            "health_check_timeout": health_check_timeout,
            "health_check_interval": health_check_interval,
        }.items()
        if v is not None
    }

    async with get_redis_client(redis_url, is_cluster=redis_cluster) as redis_client:
        consumer = InferencerConsumer(
            redis_client=redis_client, endpoint_name=endpoint_name, **kwargs
        )
        await consumer.run_forever()


@click.group()
def cli():
    pass


cli.add_command(start)



import pytest

from moriarty import mock
from moriarty.matrix.job_manager.bridge.manager import BridgeManager
from moriarty.matrix.job_manager.bridge_wrapper import BridgeWrapper
from moriarty.matrix.job_manager.params import (
    InferenceJob,
    InferenceResult,
    InferenceResultStatus,
)


@pytest.fixture
def bridge_wrapper():
    bridge_manager = BridgeManager()
    bridge_manager._load_dir(mock)
    return BridgeWrapper(bridge_manager)


async def test_bridge_wrapper(bridge_wrapper: BridgeWrapper):
    bridge = "mock"

    assert bridge_wrapper.get_bridge(bridge) is not None

    assert (
        await bridge_wrapper.enqueue_job(
            bridge,
            "test",
            InferenceJob(
                inference_id="test",
                payload={"input": "test"},
                metadata={"test": "test"},
            ),
        )
    ) == "test"

    assert (
        await bridge_wrapper.dequeue_job(
            bridge,
            "test",
            lambda x: None,
        )
        == 1
    )

    assert (
        await bridge_wrapper.enqueue_job(
            bridge,
            "test",
            InferenceJob(
                inference_id="test",
                payload={"input": "test"},
                metadata={"test": "test"},
            ),
        )
        == "test"
    )

    assert await bridge_wrapper.sample_job(bridge, "test", lambda x: None) == 1

    result = InferenceResult(
        inference_id="test",
        status=InferenceResultStatus.COMPLETED,
        message="Oh yeah",
        payload='{"output": "test"}',
    )
    assert await bridge_wrapper.enqueue_result(
        bridge,
        "test",
        result,
    )

    assert await bridge_wrapper.dequeue_result(
        bridge,
        "test",
        lambda x: None,
    ) == [result]



from __future__ import annotations

import os
from contextlib import asynccontextmanager, contextmanager
from glob import glob
from pathlib import Path
from subprocess import check_call
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, AsyncGenerator
from urllib.parse import urlparse

import alembic.command
import alembic.config
from alembic.script import ScriptDirectory
from fastapi import Depends
from sqlalchemy import create_engine, delete, exc, inspect, text
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import Session

from moriarty.log import logger
from moriarty.matrix.operator_.config import Config, get_config
from moriarty.matrix.operator_.orm import Base

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession

_here = os.path.abspath(os.path.dirname(__file__))

ALEMBIC_INI_TEMPLATE_PATH = os.path.join(_here, "alembic.ini")
ALEMBIC_DIR = os.path.join(_here, "alembic")


def write_alembic_ini(alembic_ini="alembic.ini", db_url="sqlite:///jupyterhub.sqlite"):
    """Write a complete alembic.ini from our template.

    Parameters
    ----------
    alembic_ini : str
        path to the alembic.ini file that should be written.
    db_url : str
        The SQLAlchemy database url, e.g. `sqlite:///jupyterhub.sqlite`.
    """
    with open(ALEMBIC_INI_TEMPLATE_PATH) as f:
        alembic_ini_tpl = f.read()

    with open(alembic_ini, "w") as f:
        f.write(
            alembic_ini_tpl.format(
                alembic_dir=ALEMBIC_DIR,
                # If there are any %s in the URL, they should be replaced with %%, since ConfigParser
                # by default uses %() for substitution. You'll get %s in your URL when you have usernames
                # with special chars (such as '@') that need to be URL encoded. URL Encoding is done with %s.
                # YAY for nested templates?
                db_url=str(db_url).replace("%", "%%"),
            )
        )


@contextmanager
def _temp_alembic_ini(db_url):
    """Context manager for temporary JupyterHub alembic directory

    Temporarily write an alembic.ini file for use with alembic migration scripts.

    Context manager yields alembic.ini path.

    Parameters
    ----------
    db_url : str
        The SQLAlchemy database url, e.g. `sqlite:///jupyterhub.sqlite`.

    Returns
    -------
    alembic_ini: str
        The path to the temporary alembic.ini that we have created.
        This file will be cleaned up on exit from the context manager.
    """
    with TemporaryDirectory() as td:
        alembic_ini = os.path.join(td, "alembic.ini")
        write_alembic_ini(alembic_ini, db_url)
        yield alembic_ini


@contextmanager
def chdir(path):
    """Context manager to temporarily change directory"""
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def upgrade(db_url, revision="head"):
    """Upgrade the given database to revision.

    db_url: str
        The SQLAlchemy database url, e.g. `sqlite:///jupyterhub.sqlite`.
    revision: str [default: head]
        The alembic revision to upgrade to.
    """
    with _temp_alembic_ini(db_url) as alembic_ini:
        check_call(["alembic", "-c", alembic_ini, "upgrade", revision])


class DatabaseSchemaMismatch(Exception):
    pass


def _clear_revision(engine, url):
    inspector = inspect(engine)
    if inspector.has_table("alembic_version"):
        with engine.begin() as connection:
            connection.execute(text("delete from alembic_version"))

    with _temp_alembic_ini(url) as ini:
        cfg = alembic.config.Config(ini)
        scripts = ScriptDirectory.from_config(cfg)
        old_versions_files = Path(scripts.versions) / "*.py"
        # Remove *.py from old_versions
        for old_versions_file in glob(old_versions_files.as_posix()):
            os.remove(old_versions_file)

        alembic.command.revision(config=cfg, autogenerate=True, message="init")


def get_db_log_url(db_url):
    urlinfo = urlparse(db_url)
    if urlinfo.password:
        # avoid logging the database password
        urlinfo = urlinfo._replace(
            netloc="{}:[redacted]@{}:{}".format(urlinfo.username, urlinfo.hostname, urlinfo.port)
        )
        db_log_url = urlinfo.geturl()
    else:
        db_log_url = db_url
    return db_log_url


def upgrade_in_place(db_url):
    """This is a dark magic function that upgrades the database in-place."""
    # run check-db-revision first

    db_log_url = get_db_log_url(db_url)
    logger.info(f"Initializing database: {db_log_url}")
    engine = create_engine(db_url)
    Base.metadata.create_all(engine)

    with chdir(_here):
        _clear_revision(engine, db_url)
        logger.info(f"Upgrading database: {db_log_url}")
        upgrade(db_url)


def drop_all_data(db_url):
    db_log_url = get_db_log_url(db_url)
    logger.info(f"Dropping database: {db_log_url}")
    engine = create_engine(db_url)
    for t in Base.metadata.tables.values():
        logger.info(f"Delete table: {t}")
        with Session(engine) as session:
            session.execute(delete(t))
            session.commit()


def get_db_url(
    config: Config,
    async_mode=True,
):
    if async_mode:
        return f"postgresql+asyncpg://{config.db.get_db_url()}"
    else:
        return f"postgresql+psycopg2://{config.db.get_db_url()}"


async def get_db_session(
    config: Config = Depends(get_config),
) -> AsyncGenerator[AsyncSession, None]:
    """
    For fastapi dependency injection
    """
    engine = create_async_engine(get_db_url(config, async_mode=True))
    factory = async_sessionmaker(engine)
    async with factory() as session:
        try:
            yield session
            await session.commit()
        except exc.SQLAlchemyError as error:
            await session.rollback()
            raise


@asynccontextmanager
async def open_db_session(
    config: Config,
) -> AsyncGenerator[AsyncSession, None]:
    engine = create_async_engine(get_db_url(config, async_mode=True))
    factory = async_sessionmaker(engine)
    async with factory() as session:
        try:
            yield session
            await session.commit()
        except exc.SQLAlchemyError as error:
            await session.rollback()
            raise



from __future__ import annotations

import enum
from typing import Optional

from pydantic import BaseModel

from moriarty.sidecar.params import InferenceProxyStatus, MatrixCallback


class InferenceJob(BaseModel):
    inference_id: str
    payload: dict
    """
    Request payload
    """
    metadata: dict = dict()


class InferenceResultStatus(enum.Enum):
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class InferenceResult(BaseModel):
    inference_id: str
    status: InferenceResultStatus
    message: Optional[str] = None
    payload: Optional[str] = None
    """
    Response payload
    """

    @classmethod
    def from_proxy_callback(cls, callback: MatrixCallback) -> InferenceResult:
        return InferenceResult(
            inference_id=callback.inference_id,
            status=(
                InferenceResultStatus.COMPLETED
                if callback.status == InferenceProxyStatus.FINISHED
                else InferenceResultStatus.FAILED
            ),
            message=callback.msg,
            payload=callback.payload,
        )



from pathlib import Path

from moriarty.matrix.operator_.params import CreateEndpointParams

_HERE = Path(__file__).parent

if __name__ == "__main__":
    define = CreateEndpointParams(
        endpoint_name="hello-world", image="wh1isper/moriarty-compute-image:latest"
    )
    dump_path = _HERE / "config.json"
    dump_path.write_text(define.model_dump_json(indent=4))
    print(f"Dumped config to {dump_path.as_posix()}")



from __future__ import annotations

from typing import TYPE_CHECKING

import pluggy

from moriarty.envs import *

if TYPE_CHECKING:
    from moriarty.matrix.operator_.orm import EndpointORM
    from moriarty.matrix.operator_.params import EndpointRuntimeInfo


project_name = "moriarty.matrix.spawner"
"""
The entry-point name of this extension.

Should be used in ``pyproject.toml`` as ``[project.entry-points."{project_name}"]``
"""
hookimpl = pluggy.HookimplMarker(project_name)
"""
Hookimpl marker for this extension, extension module should use this marker

Example:

    .. code-block:: python

        @hookimpl
        def register(manager):
            ...
"""

hookspec = pluggy.HookspecMarker(project_name)


@hookspec
def register(manager):
    """
    For more information about this function, please check the :ref:`manager`

    We provided an example package for you in ``{project_root}/example/extension/custom-spawner``.

    Example:

    .. code-block:: python

        class CustomSpawner(Spawner):
            register_name = "example"

        from moriarty.matrix.operator_.spawner.plugin import hookimpl

        @hookimpl
        def register(manager):
            manager.register(CustomSpawner)


    Config ``project.entry-points`` so that we can find it

    .. code-block:: toml

        [project.entry-points."moriarty.matrix.spawner"]
        {whatever-name} = "{package}.{path}.{to}.{file-with-hookimpl-function}"


    You can verify it by `moriarty-matrix list-plugins`.
    """


class EnvironmentBuilder:
    model_s3_access_key_id = os.getenv(MODEL_S3_ACCESS_KEY_ID_ENV) or os.getenv("AWS_ACCESS_KEY_ID")
    model_s3_secret_access_key = os.getenv(MODEL_S3_SECRET_ACCESS_KEY_ENV) or os.getenv(
        "AWS_SECRET_ACCESS_KEY"
    )
    model_s3_endpoint_url = os.getenv(MODEL_S3_ENDPOINT_URL_ENV) or os.getenv("AWS_S3_ENDPOINT_URL")
    model_aws_region_name = os.getenv(MODEL_AWS_REGION_ENV) or os.getenv("AWS_REGION_NAME")

    def build_sidecar_environment(self, endpoint_orm: "EndpointORM") -> dict[str, str]:
        return {
            **{
                k: str(v)
                for k, v in {
                    # For brq
                    REDIS_URL_ENV: os.getenv(REDIS_URL_ENV),
                    REDIS_HOST_ENV: os.getenv(REDIS_HOST_ENV),
                    REDIS_PORT_ENV: os.getenv(REDIS_PORT_ENV),
                    REDIS_DB_ENV: os.getenv(REDIS_DB_ENV),
                    REDIS_CLUSTER_ENV: os.getenv(REDIS_CLUSTER_ENV),
                    REDIS_TLS_ENV: os.getenv(REDIS_TLS_ENV),
                    REDIS_USERNAME_ENV: os.getenv(REDIS_USERNAME_ENV),
                    REDIS_PASSWORD_ENV: os.getenv(REDIS_PASSWORD_ENV),
                    # For sidecar
                    ENDPOINT_NAME_ENV: endpoint_orm.endpoint_name,
                    INVOKE_HOST_ENV: "localhost",
                    INVOKE_PORT_ENV: endpoint_orm.invoke_port,
                    INVOKE_PATH_ENV: endpoint_orm.invoke_path,
                    HEALTH_CHECK_PATH_ENV: endpoint_orm.health_check_path,
                    ENABLE_RETRY_ENV: endpoint_orm.allow_retry,
                    CONCURRENCY_ENV: endpoint_orm.concurrency,
                    PROCESS_TIMEOUT_ENV: endpoint_orm.process_timeout,
                    HEALTH_CHECK_TIMEOUT_ENV: endpoint_orm.health_check_timeout,
                    HEALTH_CHECK_INTERVAL_ENV: endpoint_orm.health_check_interval,
                }.items()
                if v is not None
            },
        }

    def build_compute_environment(self, endpoint_orm: "EndpointORM") -> dict[str, str]:
        return {
            **{str(k): str(v) for k, v in endpoint_orm.environment_variables.items()},
        }

    def build_init_environment(self) -> dict[str, str]:
        return {
            **{
                k: str(v)
                for k, v in {
                    "AWS_ACCESS_KEY_ID": self.model_s3_access_key_id,
                    "AWS_SECRET_ACCESS_KEY": self.model_s3_secret_access_key,
                    "AWS_S3_ENDPOINT_URL": self.model_s3_endpoint_url,
                    "AWS_REGION_NAME": self.model_aws_region_name,
                    "S3_ENDPOINT_URL": self.model_s3_endpoint_url,
                }.items()
                if v is not None
            },
        }


class Spawner:
    register_name: str

    async def prepare(self) -> None:
        pass

    async def count_avaliable_instances(self, endpoint_name: str) -> int:
        raise NotImplementedError

    async def create(self, endpoint_orm: "EndpointORM") -> None:
        raise NotImplementedError

    async def update(self, endpoint_orm: "EndpointORM", need_restart: bool = True) -> None:
        raise NotImplementedError

    async def delete(self, endpoint_name: str) -> None:
        raise NotImplementedError

    async def get_runtime_info(self, endpoint_name: str) -> "EndpointRuntimeInfo":
        raise NotImplementedError

    async def scale(self, endpoint_name: str, target_replicas: int) -> None:
        raise NotImplementedError



from __future__ import annotations

import socket
import time
from typing import TYPE_CHECKING

import pytest
import redis
from click.testing import CliRunner
from fastapi.testclient import TestClient
from sqlalchemy import create_engine, exc, select, text
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import sessionmaker

import docker
from moriarty import mock
from moriarty.matrix.job_manager.bridge.manager import BridgeManager
from moriarty.matrix.job_manager.bridge_wrapper import BridgeWrapper
from moriarty.matrix.operator_.cli import drop, init
from moriarty.matrix.operator_.config import get_config
from moriarty.matrix.operator_.dbutils import get_db_url
from moriarty.matrix.operator_.operator_ import Bridger
from moriarty.matrix.operator_.spawner.manager import SpawnerManager, get_spawner

if TYPE_CHECKING:
    from docker import DockerClient


def get_port():
    # Get an unoccupied port
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def docker_client():
    try:
        client = docker.from_env()
        client.ping()
        return client
    except:
        pytest.skip("Docker is not available")


@pytest.fixture(scope="session")
def redis_port(docker_client: DockerClient):
    """
    Start a redis container and return the port
    """
    redis_port = get_port()
    container = None
    try:
        container = docker_client.containers.run(
            "redis",
            detach=True,
            ports={"6379": redis_port},
            remove=True,
        )
        time.sleep(1)  # Wait for the server to start
        while True:
            try:
                # Ping redis
                redis_client = redis.Redis(host="localhost", port=redis_port)
                redis_client.ping()
            except:
                time.sleep(0.5)
            else:
                break
        yield redis_port
    finally:
        if container:
            container.stop()


@pytest.fixture(scope="session")
def pg_port(docker_client: DockerClient):
    pg_port = get_port()
    container = None
    try:
        container = docker_client.containers.run(
            "postgres:14",
            detach=True,
            ports={"5432": pg_port},
            remove=True,
            environment={
                "POSTGRES_USER": "postgres",
                "POSTGRES_PASSWORD": "Th3PAssW0rd!1.",
                "POSTGRES_DB": "moriarty-matrix",
            },
        )
        while True:
            # Execute `pg_isready -U postgres` in the container
            try:
                # Pg is ready
                r = container.exec_run("pg_isready -U postgres")
                assert r.exit_code == 0
                assert b"accepting connections" in r.output
                # Try to connect db
                engine = create_engine(
                    f"postgresql://postgres:Th3PAssW0rd!1.@localhost:{pg_port}/moriarty-matrix"
                )
                with engine.connect() as conn:
                    conn.execute(text("SELECT 1"))
            except Exception as e:
                time.sleep(0.5)
            else:
                break
        runner = CliRunner()
        result = runner.invoke(init, env={"DB_PORT": str(pg_port)})
        assert result.exit_code == 0
        yield pg_port
    finally:
        if container:
            container.stop()


@pytest.fixture
def spawner_manager():
    spawner_manager = SpawnerManager()
    spawner_manager._load_dir(mock)

    yield spawner_manager


@pytest.fixture
def bridge_manager():
    bridge_manager = BridgeManager()
    bridge_manager._load_dir(mock)

    yield bridge_manager


@pytest.fixture
async def async_redis_client(redis_port, monkeypatch):
    monkeypatch.setenv("REDIS_PORT", str(redis_port))
    _config = get_config()

    redis_url = _config.redis.get_redis_url()
    if _config.redis.cluster:
        redis_client = redis.asyncio.RedisCluster.from_url(redis_url, decode_responses=True)
    else:
        redis_client = redis.asyncio.from_url(redis_url, decode_responses=True)
    try:
        yield redis_client
    finally:
        await redis_client.aclose()


@pytest.fixture
async def async_session(pg_port, monkeypatch):
    monkeypatch.setenv("DB_PORT", str(pg_port))
    _config = get_config()
    engine = create_async_engine(get_db_url(_config, async_mode=True))
    factory = async_sessionmaker(engine)
    async with factory() as session:
        try:
            yield session
            await session.commit()
        except exc.SQLAlchemyError as error:
            await session.rollback()
            raise


@pytest.fixture
def session(pg_port, monkeypatch):
    monkeypatch.setenv("DB_PORT", str(pg_port))
    _config = get_config()
    engine = create_engine(get_db_url(_config, async_mode=False))
    factory = sessionmaker(engine)
    with factory() as session:
        try:
            yield session
            session.commit()
        except exc.SQLAlchemyError as error:
            session.rollback()
            raise


@pytest.fixture
def bridge_wrapper(bridge_manager):
    return BridgeWrapper(bridge_manager)


@pytest.fixture
async def spawner(spawner_manager):
    return await get_spawner(
        spawner_name="mock",
        spawner_manager=spawner_manager,
    )


@pytest.fixture
async def bridger(
    spawner,
    async_redis_client: redis.asyncio.Redis,
    async_session: AsyncSession,
    bridge_wrapper: BridgeWrapper,
):
    return Bridger(
        spawner=spawner,
        bridge_name="mock",
        bridge_wrapper=bridge_wrapper,
        redis_client=async_redis_client,
        session=async_session,
    )



import enum
import uuid
from secrets import token_urlsafe

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Enum,
    Float,
    ForeignKeyConstraint,
    Identity,
    Integer,
    String,
    Text,
    Unicode,
    UniqueConstraint,
    func,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import declarative_base

from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.sidecar.params import InferenceProxyStatus

Base = declarative_base()


class EndpointORM(Base):
    __tablename__ = "moriarty_endpoints"

    id_ = Column(Integer, autoincrement=True, primary_key=True)

    # Medatata
    endpoint_name = Column(String(255), unique=True, nullable=False, index=True)
    created_at = Column(DateTime, nullable=False, default=func.now())
    updated_at = Column(
        DateTime,
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
        comment="Updated at",
    )
    queue_capacity = Column(Integer, nullable=False, default=5)
    image = Column(String(255), nullable=True, default="wh1isper/moriarty-compute-image:latest")
    model_path = Column(
        Text,
        nullable=True,
        comment="Model path for init container to download from",
    )
    replicas = Column(
        Integer,
        nullable=False,
        default=1,
        comment="Number of replicas, autoscaler will also change it if enabled",
    )

    # Resource spec
    cpu_request = Column(Float, nullable=False, default=0.1, comment="CPU request in cores")
    cpu_limit = Column(Float, nullable=True, comment="CPU limit in milli")
    memory_request = Column(Float, nullable=False, default=128.0, comment="Memory request in MB")
    memory_limit = Column(Float, nullable=True, comment="Memory limit in MB")
    gpu_nums = Column(Integer, nullable=False, default=0)
    gpu_type = Column(String(255), nullable=False, default="nvidia.com/gpu")

    # Config for container
    environment_variables = Column(JSON, nullable=False, default={})
    environment_variables_secret_refs = Column(JSON, nullable=False, default=[])
    commands = Column(JSON, nullable=False, default=[])
    args = Column(JSON, nullable=False, default=[])
    invoke_port = Column(Integer, nullable=False, default=8080)
    invoke_path = Column(String(255), nullable=False, default="/invocations")
    health_check_path = Column(String(255), nullable=False, default="/ping")
    allow_retry = Column(Boolean, nullable=False, default=False)
    # Schedule
    node_labels = Column(JSON, nullable=False, default={})
    node_affinity = Column(JSON, nullable=False, default={})

    # Config for sidecar
    concurrency = Column(Integer, nullable=False, default=1)
    process_timeout = Column(Integer, nullable=False, default=3600)
    health_check_timeout = Column(Integer, nullable=False, default=1200)
    health_check_interval = Column(Integer, nullable=False, default=5)


class InferenceLogORM(Base):
    __tablename__ = "moriarty_inference_logs"
    id_ = Column(Integer, autoincrement=True, primary_key=True)
    inference_id = Column(String(255), nullable=False, index=True, default=lambda: uuid.uuid4().hex)
    endpoint_name = Column(String(255), nullable=False, index=True)
    inference_job = Column(JSON, nullable=False)

    status = Column(String(255), nullable=False, default=InferenceProxyStatus.RUNNING)
    callback_response = Column(JSON, nullable=True)
    finished_at = Column(DateTime, nullable=True)

    created_at = Column(DateTime, nullable=False, default=func.now())
    updated_at = Column(
        DateTime,
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
        comment="Updated at",
    )


class AutoscalerORM(Base):
    __tablename__ = "moriarty_autoscalers"

    id_ = Column(Integer, autoincrement=True, primary_key=True)
    endpoint_name = Column(String(255), unique=True, nullable=False, index=True)
    min_replicas = Column(Integer, nullable=False, default=1)
    max_replicas = Column(Integer, nullable=False, default=1)
    scale_in_cooldown = Column(Integer, nullable=False, default=60)
    scale_out_cooldown = Column(Integer, nullable=False, default=60)

    metrics = Column(String, nullable=False, default=MetricType.pending_jobs_per_instance)
    metrics_threshold = Column(Float, nullable=False, default=3)

    created_at = Column(DateTime, nullable=False, default=func.now())
    updated_at = Column(
        DateTime,
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
        comment="Updated at",
    )


class AutoscaleLogORM(Base):
    __tablename__ = "moriarty_autoscale_logs"

    id_ = Column(Integer, autoincrement=True, primary_key=True)
    endpoint_name = Column(String(255), nullable=False, index=True)
    old_replicas = Column(Integer, nullable=False)
    new_replicas = Column(Integer, nullable=False)
    metrics = Column(String, nullable=False)
    metrics_threshold = Column(Float, nullable=False)
    metrics_log = Column(JSON, nullable=False)
    details = Column(String, nullable=True)
    created_at = Column(DateTime, nullable=False, default=func.now())
    updated_at = Column(
        DateTime,
        nullable=False,
        server_default=func.now(),
        onupdate=func.now(),
        comment="Updated at",
    )



from __future__ import annotations

import asyncio
import functools
import inspect
import random
from functools import wraps

import anyio
from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def ensure_awaitable(func):
    if inspect.iscoroutinefunction(func):
        return func

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        nonlocal func
        if kwargs:
            func = functools.partial(func, **kwargs)
        return await anyio.to_thread.run_sync(func, *args)

    return wrapper


def sample_as_weights(available_priorities: list[int]) -> int:
    total = sum(available_priorities)
    weights = [p / total for p in available_priorities]
    chosen_priority = random.choices(available_priorities, weights)[0]
    return chosen_priority


class FlexibleModel(BaseModel):
    """
    This allow us provide camel and snake case in the same time
    """

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
        protected_namespaces=(),
    )



from __future__ import annotations

import math
from functools import cached_property

import redis.asyncio as redis
from fastapi import Depends, HTTPException, status
from pydantic import BaseModel
from sqlalchemy import delete, func, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from moriarty.log import logger
from moriarty.matrix.operator_.dbutils import get_db_session
from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.matrix.operator_.mixin import (
    AutoscaleMixin,
    EndpointMetrics,
    EndpointMixin,
    MetricsManager,
)
from moriarty.matrix.operator_.orm import AutoscaleLogORM, AutoscalerORM, EndpointORM
from moriarty.matrix.operator_.params import (
    AutoscaleLog,
    QueryEndpointAutoscaleLogResponse,
    QueryEndpointAutoscaleResponse,
    SetAutoscaleParams,
)
from moriarty.matrix.operator_.rds import get_redis_client
from moriarty.matrix.operator_.spawner import plugin
from moriarty.matrix.operator_.spawner.manager import get_spawner
from moriarty.sidecar.producer import JobProducer


def get_autoscaler_manager(
    spawner=Depends(get_spawner),
    redis_client=Depends(get_redis_client),
    session=Depends(get_db_session),
):
    return AutoscalerManager(redis_client=redis_client, session=session, spawner=spawner)


class CooldownMixin(AutoscaleMixin):
    prefix = "moriarty:autoscaler:cooldown"
    separator = ":"
    redis_client: redis.Redis | redis.RedisCluster

    def get_cooldown_key(self, endpoint_name: str, is_scale_out: bool) -> str:
        return self.separator.join(
            [
                self.prefix,
                "{%s}" % endpoint_name,
                "scaleup" if is_scale_out else "scaledown",
            ]
        )

    async def set_cooldown(self, endpoint_name: str, is_scale_out: bool) -> None:
        autoscaler_orm = await self.get_autoscaler_orm(endpoint_name)

        if not autoscaler_orm:
            logger.warning(f"Autoscaler not found: {endpoint_name}, may be deleted?")
            return

        cooldown_key = self.get_cooldown_key(endpoint_name, is_scale_out)
        cooldown_ttl = (
            autoscaler_orm.scale_out_cooldown if is_scale_out else autoscaler_orm.scale_in_cooldown
        )
        await self.redis_client.set(cooldown_key, 1, ex=cooldown_ttl)

    async def _is_cooldown(self, endpoint_name: str, target_replicas: int) -> bool:
        endpoint_orm = await self.get_endpoint_orm(endpoint_name)

        if not endpoint_orm or endpoint_orm.replicas == target_replicas:
            # Not exist or already scaled, treat as cooldown to skip
            return True

        cooldown_key = self.get_cooldown_key(endpoint_name, endpoint_orm.replicas < target_replicas)
        return bool(await self.redis_client.get(cooldown_key))

    async def clear_cooldown(self, endpoint_name: str) -> None:
        scaleout_key = self.get_cooldown_key(endpoint_name, True)
        scalein_key = self.get_cooldown_key(endpoint_name, False)

        await self.redis_client.delete(scaleout_key, scalein_key)


class AutoscalerManager(EndpointMixin, CooldownMixin):
    def __init__(
        self,
        redis_client: redis.Redis | redis.RedisCluster,
        session: AsyncSession,
        spawner: plugin.Spawner,
    ) -> None:
        self.redis_client = redis_client
        self.session = session
        self.spawner = spawner
        self.metrics_manager = MetricsManager(spawner=spawner, redis_client=redis_client)

    async def _clean_not_exist_endpoint(self) -> None:
        logger.debug("Cleaning not exist endpoint...")
        await self.session.execute(
            delete(AutoscalerORM).where(
                AutoscalerORM.endpoint_name.not_in(select(EndpointORM.endpoint_name))
            )
        )
        await self.session.commit()
        logger.debug("Cleaned not exist endpoint")

    async def _scale(self, endpoint_name: str) -> None:
        autoscaler_orm = await self.get_autoscaler_orm(endpoint_name)
        if not autoscaler_orm:
            logger.warning(f"Autoscaler not found: {endpoint_name}, may be deleted?")
            return

        metrics = await self.metrics_manager.get_metrics(endpoint_name)
        target_replicas = await self._calculate_target_replicas(endpoint_name, metrics)
        if target_replicas is None or await self._is_cooldown(endpoint_name, target_replicas):
            return
        try:
            logger.info(
                f"Scale endpoint `{endpoint_name}` -> {target_replicas}, based on {metrics}"
            )
            await self.spawner.scale(endpoint_name, target_replicas)
        except Exception as e:
            logger.exception(e)
            await self._revoke_scale(endpoint_name)
        else:
            await self._commit_scale(endpoint_name, metrics, target_replicas)

    async def _revoke_scale(self, endpoint_name: str) -> None:
        logger.info(f"Revoke scale endpoint `{endpoint_name}`")
        endpoint_orm = await self.get_endpoint_orm(endpoint_name)
        if endpoint_orm is None:
            logger.warning(f"Endpoint not found: {endpoint_name}, may be deleted?")
            return
        await self.spawner.scale(endpoint_name, endpoint_orm.replicas)

    async def _commit_scale(
        self, endpoint_name: str, metrics: EndpointMetrics, target_replicas: int
    ) -> None:
        endpoint_orm = await self.get_endpoint_orm(endpoint_name)
        autoscaler_orm = await self.get_autoscaler_orm(endpoint_name)

        is_scale_out = endpoint_orm.replicas < target_replicas

        logger.info(
            f"Commit scale endpoint `{endpoint_name}` {endpoint_orm.replicas} -> {target_replicas}"
        )
        log_orm = AutoscaleLogORM(
            endpoint_name=endpoint_name,
            old_replicas=endpoint_orm.replicas,
            new_replicas=target_replicas,
            metrics=autoscaler_orm.metrics,
            metrics_threshold=autoscaler_orm.metrics_threshold,
            metrics_log=metrics.model_dump(),
        )
        self.session.add(log_orm)
        await self.session.commit()
        await self.session.execute(
            update(EndpointORM)
            .where(EndpointORM.endpoint_name == endpoint_name)
            .values(replicas=target_replicas)
        )
        await self.session.commit()
        await self.set_cooldown(endpoint_name, is_scale_out)

    async def _calculate_target_replicas(
        self,
        endpoint_name: str,
        current_metrics: EndpointMetrics,
    ) -> int:
        autoscaler_orm = await self.get_autoscaler_orm(endpoint_name)

        if not autoscaler_orm:
            logger.warning(f"Autoscaler not found: {endpoint_name}, may be deleted?")
            return None

        current_metric = current_metrics.metrics.get(autoscaler_orm.metrics)
        if current_metric is None:
            logger.info(
                f"No metrics found for {autoscaler_orm.endpoint_name}.{autoscaler_orm.metrics}"
            )
            return None

        least_replicas = self.metrics_manager.calculate_least_replicas(
            metric=autoscaler_orm.metrics,
            current_metric_value=current_metric,
            metric_threshold=autoscaler_orm.metrics_threshold,
        )
        logger.info(
            f"Least replicas calculated for {autoscaler_orm.endpoint_name}({autoscaler_orm.metrics}) is {least_replicas}"
        )
        if least_replicas < autoscaler_orm.min_replicas:
            logger.info(
                f"Using min replicas {autoscaler_orm.min_replicas} instead of {least_replicas}"
            )
            return autoscaler_orm.min_replicas
        if least_replicas > autoscaler_orm.max_replicas:
            logger.info(
                f"Using max replicas {autoscaler_orm.max_replicas} instead of {least_replicas}"
            )
            return autoscaler_orm.max_replicas
        return least_replicas

    async def get_scaleable_endpoints(self) -> list[str]:
        return (
            (
                await self.session.execute(
                    select(EndpointORM.endpoint_name).where(
                        EndpointORM.endpoint_name.in_(select(AutoscalerORM.endpoint_name))
                    )
                )
            )
            .scalars()
            .all()
        )

    async def scan_and_scale(self) -> None:
        await self._clean_not_exist_endpoint()
        for endpoint in await self.get_scaleable_endpoints():
            await self._scale(endpoint)

        await self.session.commit()

    async def get_autoscale_info(self, endpoint_name: str) -> QueryEndpointAutoscaleResponse:
        autoscaler_orm = await self.get_autoscaler_orm(endpoint_name)
        if autoscaler_orm is None:
            raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Autoscaler not found")

        return QueryEndpointAutoscaleResponse(
            endpoint_name=endpoint_name,
            min_replicas=autoscaler_orm.min_replicas,
            max_replicas=autoscaler_orm.max_replicas,
            scale_in_cooldown=autoscaler_orm.scale_in_cooldown,
            scale_out_cooldown=autoscaler_orm.scale_out_cooldown,
            metrics=autoscaler_orm.metrics,
            metrics_threshold=autoscaler_orm.metrics_threshold,
        )

    async def set_autoscale(
        self, endpoint_name: str, params: SetAutoscaleParams
    ) -> QueryEndpointAutoscaleResponse:
        endpoint_orm = await self.get_endpoint_orm(endpoint_name)
        if endpoint_orm is None:
            raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Endpoint not found")

        params = {
            k: v
            for k, v in dict(
                min_replicas=params.min_replicas,
                max_replicas=params.max_replicas,
                scale_in_cooldown=params.scale_in_cooldown,
                scale_out_cooldown=params.scale_out_cooldown,
                metrics=params.metrics,
                metrics_threshold=params.metrics_threshold,
            ).items()
            if v is not None
        }
        if not await self.get_autoscaler_orm(endpoint_name):
            autoscaler_orm = AutoscalerORM(
                endpoint_name=endpoint_name,
                **params,
            )
            self.session.add(autoscaler_orm)
        else:
            await self.session.execute(
                update(AutoscalerORM)
                .where(AutoscalerORM.endpoint_name == endpoint_name)
                .values(params)
            )

        await self.session.commit()
        await self.clear_cooldown(endpoint_name)

        logger.info(f"Autoscaler {endpoint_name} updated")
        return await self.get_autoscale_info(endpoint_name)

    async def delete(self, endpoint_name: str) -> None:
        await self.session.execute(
            delete(AutoscalerORM).where(AutoscalerORM.endpoint_name == endpoint_name)
        )

    async def get_autoscale_logs(
        self,
        endpoint_name: str,
        limit: int,
        cursor: int | None,
        order_by: str,
        order: str,
    ) -> QueryEndpointAutoscaleLogResponse:
        if cursor:
            cursor = await self.session.get(AutoscaleLogORM, cursor)
            if not cursor:
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND, detail="Cursor not found"
                )

        query = select(AutoscaleLogORM).where(AutoscaleLogORM.endpoint_name == endpoint_name)

        if cursor:
            query = query.where(AutoscaleLogORM.id_ < cursor.id_)

        if order == "desc":
            query = query.order_by(getattr(AutoscaleLogORM, order_by).desc())

        else:
            query = query.order_by(getattr(AutoscaleLogORM, order_by))

        query = query.limit(limit)
        total = (
            await self.session.execute(select(func.count(AutoscaleLogORM.endpoint_name)))
        ).scalar_one()
        log_orms = (await self.session.execute(query)).scalars().all()
        next_cursor = log_orms[-1].id_ if log_orms else None

        return QueryEndpointAutoscaleLogResponse(
            logs=[
                AutoscaleLog(
                    id_=log_orm.id_,
                    endpoint_name=log_orm.endpoint_name,
                    old_replicas=log_orm.old_replicas,
                    new_replicas=log_orm.new_replicas,
                    metrics=log_orm.metrics,
                    metrics_threshold=log_orm.metrics_threshold,
                    metrics_log=log_orm.metrics_log,
                    details=log_orm.details,
                    created_at=log_orm.created_at,
                    updated_at=log_orm.updated_at,
                )
                for log_orm in log_orms
                if log_orm
            ],
            cursor=next_cursor,
            limit=limit,
            total=total,
        )



import os
from contextlib import asynccontextmanager

from fastapi import Depends, FastAPI, HTTPException, Query, Response, status

from moriarty.envs import MORIARTY_MATRIX_TOKEN_ENV
from moriarty.log import logger
from moriarty.matrix.operator_.autoscaler import (
    AutoscalerManager,
    get_autoscaler_manager,
)
from moriarty.matrix.operator_.operator_ import Operator, get_operaotr
from moriarty.matrix.operator_.params import (
    CreateEndpointParams,
    ListEndpointsResponse,
    QueryEndpointAutoscaleLogResponse,
    QueryEndpointAutoscaleResponse,
    QueryEndpointResponse,
    SetAutoscaleParams,
    UpdateEndpointParams,
)

TOKEN = os.getenv(MORIARTY_MATRIX_TOKEN_ENV)


@asynccontextmanager
async def lifespan(app: FastAPI):
    if TOKEN:
        logger.info(f"Token: {TOKEN}")
    else:
        logger.info("Token not set, authentication is disabled")
    yield


app = FastAPI(lifespan=lifespan)


@app.middleware("http")
async def verify_token(request, call_next):
    if request.url.path == "/":
        return await call_next(request)

    if TOKEN and request.headers.get("Authorization") != f"Bearer {TOKEN}":
        return Response(
            status_code=401,
            content=f"Unauthorized. Check environment {MORIARTY_MATRIX_TOKEN_ENV} for token.",
        )
    return await call_next(request)


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.get("/endpoint/list")
async def list_endpoints(
    limit: int = Query(100, ge=1, le=1000),
    cursor: str | None = None,
    keyword: str | None = None,
    orderBy: str = Query("created_at", pattern="^(created_at|updated_at|endpoint_name)$"),
    order: str = Query("desc", pattern="^(asc|desc)$"),
    operator: Operator = Depends(get_operaotr),
) -> ListEndpointsResponse:
    return await operator.list_endpoints(
        limit=limit,
        cursor=cursor,
        keyword=keyword,
        order_by=orderBy,
        order=order,
    )


@app.post("/endpoint/create")
async def create_endpoint(
    params: CreateEndpointParams,
    operaotr: Operator = Depends(get_operaotr),
):
    await operaotr.create_endpoint(params)
    return Response(status_code=status.HTTP_201_CREATED)


@app.get("/endpoint/{endpoint_name}/info")
async def get_endpoint_info(
    endpoint_name: str,
    operator: Operator = Depends(get_operaotr),
) -> QueryEndpointResponse:
    return await operator.get_endpoint_info(endpoint_name)


@app.post("/endpoint/{endpoint_name}/update")
async def update_endpoint(
    endpoint_name: str,
    params: UpdateEndpointParams,
    operator: Operator = Depends(get_operaotr),
) -> QueryEndpointResponse:
    return await operator.update_endpoint(endpoint_name, params)


@app.post("/endpoint/{endpoint_name}/delete")
async def delete_endpoint(
    endpoint_name: str,
    operator: Operator = Depends(get_operaotr),
):
    await operator.delete_endpoint(endpoint_name)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@app.get("/autoscale/{endpoint_name}/info")
async def autoscale_endpoint_info(
    endpoint_name: str,
    autoscaler: AutoscalerManager = Depends(get_autoscaler_manager),
) -> QueryEndpointAutoscaleResponse:
    return await autoscaler.get_autoscale_info(endpoint_name)


@app.post("/autoscale/{endpoint_name}/set")
async def set_autoscale(
    endpoint_name: str,
    params: SetAutoscaleParams,
    autoscaler: AutoscalerManager = Depends(get_autoscaler_manager),
) -> QueryEndpointAutoscaleResponse:
    return await autoscaler.set_autoscale(endpoint_name, params)


@app.post("/autoscale/{endpoint_name}/delete")
async def delete_autoscale(
    endpoint_name: str,
    autoscaler: AutoscalerManager = Depends(get_autoscaler_manager),
) -> QueryEndpointAutoscaleResponse:
    await autoscaler.delete(endpoint_name)
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@app.get("/autoscale/{endpoint_name}/log")
async def get_autoscale_log(
    endpoint_name: str,
    limit: int = Query(100, ge=1, le=1000),
    cursor: int | None = None,
    orderBy: str = Query("created_at", pattern="^(created_at|updated_at|endpoint_name)$"),
    order: str = Query("desc", pattern="^(asc|desc)$"),
    autoscaler: AutoscalerManager = Depends(get_autoscaler_manager),
) -> QueryEndpointAutoscaleLogResponse:
    return await autoscaler.get_autoscale_logs(
        endpoint_name, limit=limit, cursor=cursor, order_by=orderBy, order=order
    )



from __future__ import annotations

from typing import Awaitable

from moriarty.log import logger
from moriarty.matrix.job_manager.bridge.plugin import QueueBridge, hookimpl
from moriarty.matrix.job_manager.params import (
    InferenceJob,
    InferenceResult,
    InferenceResultStatus,
)
from moriarty.utils import Singleton


class MockBridge(QueueBridge, metaclass=Singleton):
    register_name = "mock"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.bridge_result_queue_url = "result"
        self.endpoints = {}
        self.result_queue = []

    def make_job_queue_url(self, endpoint_name: str, priority: int = None) -> str:
        priority = priority or 100
        endpoint_queue_map = self.endpoints.setdefault(endpoint_name, {})
        endpoint_queue_map.setdefault(priority, [])
        return f"{endpoint_name}.{priority}"

    def remove_job_queue(self, endpoint_name: str, priority: int = None) -> None:
        priority = priority or 100
        endpoint_queue_map = self.endpoints.setdefault(endpoint_name, {})
        endpoint_queue_map[priority] = []

    async def list_avaliable_priorities(self, endpoint_name: str) -> list[int]:
        return list(
            k for k in self.endpoints[endpoint_name].keys() if self.endpoints[endpoint_name][k]
        )

    async def enqueue_job(self, endpoint_name: str, job: InferenceJob, priority: int = None) -> str:
        queue_url = self.make_job_queue_url(endpoint_name, priority=priority)
        k1, k2 = queue_url.split(".")
        self.endpoints[k1][int(k2)].append(job)
        return job.inference_id

    async def dequeue_job(
        self,
        endpoint_name: str,
        process_func: Awaitable[InferenceJob],
        size: int = 1,
        priority: int = None,
    ) -> int:
        queue_url = self.make_job_queue_url(endpoint_name, priority=priority)
        k1, k2 = queue_url.split(".")
        jobs = []
        for _ in range(size):
            if not self.endpoints[k1][int(k2)]:
                break
            jobs.append(self.endpoints[k1][int(k2)].pop(0))

        for job in jobs:
            await process_func(job)
        return len(jobs)

    async def enqueue_result(self, result: InferenceResult) -> str:
        self.result_queue.append(result)
        return result.inference_id

    async def dequeue_result(
        self, process_func: Awaitable[InferenceResult], size: int = 10
    ) -> list[InferenceResult]:
        size = size or 1
        results = []
        for _ in range(size):
            if not self.result_queue:
                break
            results.append(self.result_queue.pop(0))
        for result in results:
            await process_func(result)
        return results


@hookimpl
def register(manager):
    manager.register(MockBridge)
